/// @file       CommonFunctions.cpp
/// @brief      Common functions for all modules
/// @details    Provide miscellaneous functions for all modules 
/// @version    1.0.0
/// @date       2014.01.21
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.hpp"
#define INCLUDE_ERRORTEXT 
#include <error_code.hpp>
#undef INCLUDE_ERRORTEXT
#include "CommonFunctions.hpp"

/// @fn         deleteErrorMsg(ErrorMsg* in_pErrorMsg)
/// @brief      Delete ErrorMsg list and free its allocated memory
/// @param      in_pErrorMsg is a pointer to the list of ErrorMsg
/// @exception  Nothing
/// @return     Nothing
extern void
deleteErrorMsg(ErrorMsg* in_pErrorMsg)
{
    ErrorMsg* pCurrent = in_pErrorMsg;
    ErrorMsg* pTemp = NULL;

    while (pCurrent)
    {
        pTemp = pCurrent->getNextErrorMsg();
        delete pCurrent;
        pCurrent = pTemp;
    }
}

/// @fn         trimRightChar(CHAR* in_szString,
///                           CHAR  in_chJunk)
/// @brief      Trim junk character from right of string 
/// @param      in_szString is a char string
/// @param      in_chJunk is a character for trimming
/// @exception  Nothing
/// @return     Nothing
extern CHAR*
trimRightChar(CHAR* in_szString,
              CHAR  in_chJunk)
{
    CHAR* szOriginal = in_szString + strlen(in_szString);

    while(*--szOriginal == in_chJunk);
       *(szOriginal + 1) = '\0';

    return in_szString;
}

/// @fn         trimLeftChar(CHAR* in_szString,
///                          CHAR  in_chJunk)
/// @brief      Trim junk character from left of string 
/// @param      in_szString is a char string
/// @param      in_chJunk is a character for trimming
/// @exception  Nothing
/// @return     Nothing
extern CHAR*
trimLeftChar(CHAR *in_szString,
             CHAR in_chJunk)
{
    CHAR*   szOriginal = in_szString;
    CHAR*   chpTemp = szOriginal;
    INT32   s32Trimmed = 0;

    do
    {
        if (*szOriginal != in_chJunk || s32Trimmed)
        {
            s32Trimmed = 1;
            *chpTemp++ = *szOriginal;
        }
    }
    while (*szOriginal++ != '\0');

    return in_szString;
}

/// @fn         parseErrorText(ERRORCODE    in_eErrorCode, 
///                            UINT32       in_u32SearchPattern, 
///                            string*      out_strpMessage)
/// @brief      Translate error code to error message
/// @param      in_eErrorCode is a 32bits error code
/// @param      in_u32SearchPattern is 32bits search pattern for searching error text table
/// @param      out_strpMessage is a pointer to string for returned message
/// @exception  Nothing
/// @return     Nothing
void
parseErrorText(ERRORCODE    in_eErrorCode, 
               UINT32       in_u32SearchPattern, 
               string*      out_strpMessage)
{
    INT32   i;
    INT32   s32TableSize = sizeof(ErrorCodeTextTableCollect)/sizeof(ErrorCodeTextTable);
    ErrorCodeText* pErrorCodeTextTable = NULL;

    for (i = 0; i < s32TableSize; i++)
    {
        if (in_u32SearchPattern == ErrorCodeTextTableCollect[i].u32Index)
        {
            break;
        }
    }
        
    if (i == s32TableSize)
    {
        // FIXME: Need to replace this by common debug print function
        //printf("Unknown Search Pattern for Error Code Text Table\n");
        return;
    }
    else
    {
        pErrorCodeTextTable = ErrorCodeTextTableCollect[i].errorCodeTextTable;
        s32TableSize = ErrorCodeTextTableCollect[i].s32TableSize;
    }
    
    for (i = 0; i < s32TableSize; i++)
    {
        if (pErrorCodeTextTable[i].u32ErrorCode == in_eErrorCode)
        {
            (*out_strpMessage).append(pErrorCodeTextTable[i].strErrorText);
            break;    
        }
    }

}

/// @fn         parseErrorTextFail(string* out_strpMessage)
/// @brief      Error handling when ErrorMsg parser is failed
/// @details    Generate error message when ErrorMsg parser is failed
/// @param      out_strpMessage is a pointer to string for returned message
/// @exception  Nothing
/// @return     Nothing
void
parseErrorTextFail(string* out_strpMessage)
{
    (*out_strpMessage).append("Fail to parse error code!!");
}

/// @fn         getErrorMsgText(ErrorMsg*   in_pErrorMsg, 
///                             string*     out_strpMessage)
/// @brief      Get error message for one ErrorMsg
/// @details    Translate ErrorMsg to error messages 
/// @param      in_pErrorMsg is a pointer to one ErrorMsg
/// @param      out_strpMessage is a pointer to string for returned message
/// @exception  Nothing
/// @return     Nothing
void 
getErrorMsgText(ErrorMsg*   in_pErrorMsg, 
                string*     out_strpMessage)
{
    ERRORCODE   eErrorCode = (ERRORCODE)in_pErrorMsg->getErrorCode();
    UINT32      u32SearchPattern = 0;
    string      strModuleText;
    string      strTypeText;
    string      strAPIText;
    string      strErrorText;
    string      strTemp;
    
    // Get module error code text
    u32SearchPattern = ERRORCODE_MASK_MODULE;
    parseErrorText((ERRORCODE)(eErrorCode & u32SearchPattern), u32SearchPattern, &strModuleText);
    if (strModuleText.empty())
    {
        return parseErrorTextFail(out_strpMessage);
    }
    else
    {
        (*out_strpMessage).append(strModuleText);
    }

    // Get type error code text
    u32SearchPattern = ERRORCODE_MASK_TYPE;
    parseErrorText((ERRORCODE)(eErrorCode & u32SearchPattern), u32SearchPattern, &strTypeText);
    if (strTypeText.empty())
    {
        return parseErrorTextFail(out_strpMessage);
    }
    else
    {
        (*out_strpMessage).append(strTypeText);
    }

    // Get API error code text
    u32SearchPattern = (eErrorCode & (MEF_HEADER | ERRORCODE_MASK_MODULE | ERRORCODE_MASK_TYPE));
    if (eErrorCode & ERRORCODE_MASK_API)
    {
        parseErrorText((ERRORCODE)(eErrorCode & (u32SearchPattern | ERRORCODE_MASK_API)), u32SearchPattern, &strAPIText);
        if (strAPIText.empty())
            return parseErrorTextFail(out_strpMessage);
        else
        {
            (*out_strpMessage).append("::");
            (*out_strpMessage).append(strAPIText);
        }
    }

    // Get common or module error code text
    if (eErrorCode & ERRORCODE_MASK_ERROR)
    {
        if ((eErrorCode & ERROR_COMMON_ERROR_CODE) == ERROR_COMMON_ERROR_CODE)
            u32SearchPattern = ERROR_COMMON_ERROR_CODE;
        else
            u32SearchPattern = (eErrorCode & (MEF_HEADER | ERRORCODE_MASK_MODULE | ERRORCODE_MASK_TYPE));
    
        parseErrorText((ERRORCODE)(eErrorCode & (u32SearchPattern | ERRORCODE_MASK_ERROR)), u32SearchPattern, &strErrorText);

        if (strErrorText.empty())
        {
            return parseErrorTextFail(out_strpMessage);
        }
        else
        {
            (*out_strpMessage).append(": \"");
            (*out_strpMessage).append(strErrorText);

            // Append module specified message to error message
            strTemp = in_pErrorMsg->getDetailMessage();            
            if (!strTemp.empty())
            {
                (*out_strpMessage).append(" - ");
                (*out_strpMessage).append(strTemp);
            }
            
            (*out_strpMessage).append("\"");
        }
    }
    else
    {
        // Append module specified message to error message
        strTemp = in_pErrorMsg->getDetailMessage();            
        if (!strTemp.empty())            
        {
            (*out_strpMessage).append(": \"");
            (*out_strpMessage).append(strTemp);
            (*out_strpMessage).append("\"");
        }
    }
}

/// @fn         getErrorMessage(ErrorMsg*   in_pErrorMsgList, 
///                             string*     out_strpMessage)
/// @brief      Get error messages
/// @details    Translate ErrorMsg list to error messages 
/// @param      in_pErrorMsgList is a pointer to the error message list
/// @param      out_strpMessage is a pointer to string for returned message
/// @exception  Nothing
/// @return     Nothing
extern void 
getErrorMessage(ErrorMsg*   in_pErrorMsgList, 
                string*     out_strpMessage)
{
    ErrorMsg*   pCurrentErrorMsg = in_pErrorMsgList;
    ErrorMsg*   pTempErrorMsg = NULL;
    
    while (pCurrentErrorMsg)
    {
        if (pCurrentErrorMsg->isListHead() == FALSE)
        {
            (*out_strpMessage).append(" -> ");
        }
            
        pTempErrorMsg = pCurrentErrorMsg->getNextErrorMsg();
        getErrorMsgText(pCurrentErrorMsg, out_strpMessage);

        pCurrentErrorMsg = pTempErrorMsg;
    }

    deleteErrorMsg(in_pErrorMsgList);
}



extern void 
getLastErrorMessage(ErrorMsg*   in_pErrorMsgList, 
                    string*     out_strpMessage)
{
    string TempStress, token, ErrorResult;
    std::istringstream iss;
    //int index = 0;

    getErrorMessage(in_pErrorMsgList, &TempStress);
    iss.str(TempStress);

    while (getline(iss, token, '"'))
    {
        ErrorResult = token;
    }

    *out_strpMessage = ErrorResult;
}



extern string
Int2HexStr(UINT32 i, UINT8 digits)
{
  std::stringstream stream;
  stream << std::hex << std::setw(digits) << std::setfill('0')  << i;
  return stream.str();
}


extern string
Int2Str(INT32 i)
{
  string s, result_string;
  std::stringstream ss(s);
  ss << i;
  result_string = ss.str();
  return result_string.c_str();
}

/// @fn         UINT64ToHexStr(UINT64 in_u64Value)
/// @brief      Translate UINT64 value to hexdecimal string 
/// @param      in_u64Value is a 64 bits value 
/// @exception  Nothing
/// @return     A string for hexdecimal value
extern string   
UINT64ToHexStr(UINT64 in_u64Value)
{
    std::stringstream stream;
    stream << "0x" << std::hex << std::setw(16) << std::setfill('0')  << in_u64Value;
    return stream.str();
}

/// @fn         BCDToStr(UINT8 in_u8BCDValue)
/// @brief      Translate BCD 8 bits value to string
/// @param      in_u8BCDValue is a 8 bits value 
/// @exception  Nothing
/// @return     A string for BCD value
extern string
BCDToStr(UINT8 in_u8BCDValue)
{
    string strValue;

    strValue = std::to_string((UINT32)(in_u8BCDValue & 0xF0)) + \
               std::to_string((UINT32)(in_u8BCDValue & 0x0F));

    return strValue;
}

extern ARCH_TYPE
getArch(void)
{
    FILE*   pFile = NULL;
    CHAR    szArch[8];
    
    // Get architecture
    memset(szArch, '\0', sizeof(szArch));
    
    if ((pFile = popen("uname -r", "r")) != NULL)
    {
        fgets(szArch, sizeof(szArch), pFile);
        
        pclose(pFile); 
    }

    if (strstr(szArch, "arm") != NULL)
    {
        return ARCH_ARM;
    }
    else if (strstr(szArch, "ppc") != NULL)
    {
        return ARCH_PPC;
    }
    else
    {
        return ARCH_X86;
    }
}



extern void
ConvertCharToUpperCase(CHAR* userCharacter)
{
    *userCharacter = toupper(*userCharacter);
}



extern void
ConvertStringToUpperCase(CHAR* userString)
{
    CHAR*  pointer;

    pointer = userString;
    while (*pointer != '\0')
    {
        *pointer = toupper(*pointer);
        pointer++;
    }
}


extern void
ConvertCharToLowerCase(CHAR* userCharacter)
{
    *userCharacter = tolower(*userCharacter);
}



extern void
ConvertStringLowerCase(CHAR* userString)
{
    CHAR*  pointer;

    pointer = userString;
    while (*pointer != '\0')
    {
        *pointer = tolower(*pointer);
        pointer++;
    }
}


