/// @file       MemoryDAPI_AddressLineTest.cpp
/// @brief      Provide address line test for MemoryDAPI
/// @details    Verify memory address lines by write and read test pattern for power of 2 addressing
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "CommonFunctions.hpp"
#include "MemoryDAPI.hpp"

/// @fn         MemoryDAPI::addressLineTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                         PHYSICAL_ADDRESS_T in_endAddress,
///                                         INT32              in_s32ElementSize,
///                                         UINT64             in_u64UserTestPattern)
/// @brief      Verify memory address lines
/// @param      in_startAddress is a physcial address for start address
/// @param      in_endAddress is a physcial address for end address
/// @param      in_s32ElementSize is 32bits value for element size (ex: 8/16/32/64)
/// @param      in_u64UserTestPattern is a 64bits test pattern
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::addressLineTest(PHYSICAL_ADDRESS_T in_startAddress,
                            PHYSICAL_ADDRESS_T in_endAddress,
                            INT32              in_s32ElementSize,
                            UINT64             in_u64UserTestPattern)
{
    ErrorMsg*           pErrorMsg = NULL;
    string              strErrorMsg;
    PHYSICAL_ADDRESS_T  startAddress = in_startAddress;
    PHYSICAL_ADDRESS_T  endAddress = in_endAddress;
    PHYSICAL_ADDRESS_T  testBlockSize = 0;
    UINT32              u32Steps = 0;
    PHYSICAL_ADDRESS_T  offset = 0;
    PHYSICAL_ADDRESS_T  tempOffset = 0;
    BOOL                bIsFailed = FALSE;               

    switch (in_s32ElementSize)
    {
        case 8:
        case 16:
        case 32:
        case 64:
            // Round up memory address for alignment
            startAddress = alignMemoryAddress(startAddress, in_s32ElementSize);

            testBlockSize = endAddress - startAddress;
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_DATA_LINE_TEST | ERROR_BAD_PARAMETER_THREE), 
                                  "Unknown element size",
                                  NULL);
            break;

    }

    if (testBlockSize == 0)
    {
        return NO_ERROR_MSG;
    }

    offset = 0;   
    while (offset < testBlockSize)
    {
        if (offset != 0)
        {
            offset <<= 1;
        }
        else
        {
            offset = in_s32ElementSize >> 3;
        }

        u32Steps++;
    }
    
    initProgress(u32Steps);

    // Write test pattern for power of 2 addressing
    // Ex: For 32bits element size, 0, 4, 8, 16, ...
    offset = (in_s32ElementSize >> 3);
    
    while (offset < testBlockSize)
    {   
        switch (in_s32ElementSize)
        {
            case 8:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT8)in_u64UserTestPattern);                
                break;
            case 16:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT16)in_u64UserTestPattern);               
                break;
            case 32:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT32)in_u64UserTestPattern);                
                break;
            case 64:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT64)in_u64UserTestPattern);
                break;
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
        }

        offset <<= 1;
    }

    // Write inverse test pattern to verify remaining address
    offset = 0;
    
    while (offset < testBlockSize)
    {
        switch (in_s32ElementSize)
        {
            case 8:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT8)(~in_u64UserTestPattern));                
                break;
            case 16:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT16)(~in_u64UserTestPattern));               
                break;
            case 32:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT32)(~in_u64UserTestPattern));                
                break;
            case 64:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT64)(~in_u64UserTestPattern));
                break;
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
        }

        // Verify test pattern for remaining address
        tempOffset = 0;
        
        while (tempOffset < testBlockSize)
        {
            switch (in_s32ElementSize)
            {
                case 8:
                    if (((UINT8)in_u64UserTestPattern != m_pMemoryDDAPI->read8((startAddress + tempOffset))) 
                        && (offset != tempOffset))
                    {
                        bIsFailed = TRUE;
                    }
                    break;
                case 16:
                    if (((UINT16)in_u64UserTestPattern != m_pMemoryDDAPI->read16((startAddress + tempOffset))) 
                        && (offset != tempOffset))
                    {
                        bIsFailed = TRUE;
                    }               
                    break;
                case 32:
                    if (((UINT32)in_u64UserTestPattern != m_pMemoryDDAPI->read32((startAddress + tempOffset))) 
                        && (offset != tempOffset))
                    {
                        bIsFailed = TRUE;
                    }               
                    break;
                case 64:
                    if (((UINT64)in_u64UserTestPattern != m_pMemoryDDAPI->read64((startAddress + tempOffset))) 
                        && (offset != tempOffset))
                    {
                        bIsFailed = TRUE;
                    }
                    break;
            }

            if (bIsFailed)
            {
                initProgress(0);
                
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    strErrorMsg = "failed address: " + UINT64ToHexStr(startAddress + tempOffset);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), 
                                          strErrorMsg.c_str(),
                                          NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                }
            } 

            if (tempOffset != 0)
            {
                tempOffset <<= 1;
            }
            else
            {
                // We should align the second offset address
                tempOffset = (in_s32ElementSize >> 3);
            }
        }

        switch (in_s32ElementSize)
        {
            case 8:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT8)in_u64UserTestPattern);                
                break;
            case 16:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT16)in_u64UserTestPattern);               
                break;
            case 32:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT32)in_u64UserTestPattern);                
                break;
            case 64:
                pErrorMsg = m_pMemoryDDAPI->write((startAddress + offset), (UINT64)in_u64UserTestPattern);
                break;
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
        }   

        if (offset != 0)
        {
            offset <<= 1;
        }
        else
        {
            // We should align the second offset address
            offset = (in_s32ElementSize >> 3);
        }  

        updateProgress(1);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::addressLineTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                         UINT64             in_u64TestSize,
///                                         INT32              in_s32Controller,
///                                         INT32              in_s32Channel,
///                                         INT32              in_s32Rank,
///                                         INT32              in_s32ElementSize,
///                                         UINT64             in_u64UserTestPattern)
/// @brief      Verify memory address lines
/// @param      in_startAddress is a physical address for start address
/// @param      in_u64TestSize is an UINT64 value for test size
/// @param      in_s32Controller is an INT32 value for target controller
/// @param      in_s32Channel is an INT32 value for target channel
/// @param      in_s32Rank is an INT32 value for target rank
/// @param      in_s32ElementSize is 32bits value for element size (ex: 8/16/32/64)
/// @param      in_u64UserTestPattern is a 64bits test pattern
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::addressLineTest(PHYSICAL_ADDRESS_T in_startAddress,
                            UINT64             in_u64TestSize,
                            INT32              in_s32Controller,
                            INT32              in_s32Channel,
                            INT32              in_s32Rank,
                            INT32              in_s32ElementSize,
                            UINT64             in_u64UserTestPattern)
{
    ErrorMsg*           pErrorMsg = NULL;
    string              strErrorMsg;
    PHYSICAL_ADDRESS_T  rankOffset = 0, lastRankOffset = 0, currentAddress = 0;
    UINT64              u64DataByteWidth, u64TestStep = 0;
    PHYSICAL_ADDRESS_T  testBlockSize = 0, offset = 0, tempOffset = 0;
    BOOL                bIsFailed = FALSE;

    // Workaround: For progress status display
    initProgress(in_s32ElementSize);

    // Get the rank offset based on the start address
    if ((pErrorMsg = m_pMemoryDDAPI->getRankOffset(in_startAddress, &rankOffset)) != NO_ERROR_MSG)
    {
        initProgress(0);
        return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
    }

    switch (in_s32ElementSize)
    {
        case 8:
        case 16:
        case 32:
        case 64:
            // Round up memory address for alignment
            rankOffset = alignMemoryAddress(rankOffset, in_s32ElementSize);

            u64DataByteWidth = in_s32ElementSize >> 3;
            // Round up test size for alignment
            testBlockSize = (PHYSICAL_ADDRESS_T)((in_u64TestSize / u64DataByteWidth) * u64DataByteWidth);
            break;
        default:
            initProgress(0);
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST | ERROR_BAD_PARAMETER_THREE),
                                  "Unknown element size",
                                  NULL);
    }

    if (testBlockSize == 0)
    {
        initProgress(0);
        return NO_ERROR_MSG;
    }

    lastRankOffset = rankOffset + testBlockSize;

    if (rankOffset && (testBlockSize > rankOffset))
    {
        testBlockSize = rankOffset;
    }

    if (rankOffset != 0)
    {
        offset = 0;
        while (offset < testBlockSize)
        {
            if (offset != 0)
            {
                offset <<= 1;
            }
            else
            {
                offset = u64DataByteWidth;
            }

            u64TestStep++;
        }
    }

    offset = rankOffset << 1;
    while (offset < lastRankOffset)
    {
        if (offset != 0)
        {
            offset <<= 1;
        }
        else
        {
            offset = u64DataByteWidth;
        }

        u64TestStep++;
    }


    initProgress(u64TestStep);

    // Write test pattern for power of 2 addressing
    // Ex: For 32bits element size, 0, 4, 8, 16, ...
    if (rankOffset != 0)
    {
        offset = 0;
        while (offset < testBlockSize)
        {
            pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress((rankOffset + offset),
                                                           in_s32Controller,
                                                           in_s32Channel,
                                                           in_s32Rank,
                                                           &currentAddress);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)in_u64UserTestPattern);
                    break;
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)in_u64UserTestPattern);
                    break;
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)in_u64UserTestPattern);
                    break;
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT64)in_u64UserTestPattern);
                    break;
            }

            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            if (offset)
            {
                offset <<= 1;
            }
            else
            {
                offset = u64DataByteWidth;
            }
        }
    }

    offset = rankOffset << 1;
    while (offset < lastRankOffset)
    {
        pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(offset,
                                                       in_s32Controller,
                                                       in_s32Channel,
                                                       in_s32Rank,
                                                       &currentAddress);
        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
        }

        switch (in_s32ElementSize)
        {
            case 8:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)in_u64UserTestPattern);
                break;
            case 16:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)in_u64UserTestPattern);
                break;
            case 32:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)in_u64UserTestPattern);
                break;
            case 64:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT64)in_u64UserTestPattern);
                break;
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
        }

        if (offset)
        {
            offset <<= 1;
        }
        else
        {
            offset = u64DataByteWidth;
        }
    }

    // Write inverse test pattern to verify remaining address
    if (rankOffset != 0)
    {
        // Verify from rankOffset plus an offset which will be power of 2
        offset = 0;
        while (offset < testBlockSize)
        {
            // Write inverse test pattern
            pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress((rankOffset + offset),
                                                           in_s32Controller,
                                                           in_s32Channel,
                                                           in_s32Rank,
                                                           &currentAddress);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)(~in_u64UserTestPattern));
                    break;
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)(~in_u64UserTestPattern));
                    break;
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)(~in_u64UserTestPattern));
                    break;
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT64)(~in_u64UserTestPattern));
                    break;
            }

            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            // Check remaining addresses from rankOffset plus an offset which will be power of 2.
            tempOffset = 0;
            while (tempOffset < testBlockSize)
            {
                pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress((rankOffset + tempOffset),
                                                               in_s32Controller,
                                                               in_s32Channel,
                                                               in_s32Rank,
                                                               &currentAddress);
                if (pErrorMsg != NO_ERROR_MSG)
                {
                    initProgress(0);
                    return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                }

                switch (in_s32ElementSize)
                {
                    case 8:
                        if ((offset != tempOffset) &&
                            ((UINT8)in_u64UserTestPattern != m_pMemoryDDAPI->read8(currentAddress)))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 16:
                        if ((offset != tempOffset) &&
                            ((UINT16)in_u64UserTestPattern != m_pMemoryDDAPI->read16(currentAddress)))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 32:
                        if ((offset != tempOffset) &&
                            ((UINT32)in_u64UserTestPattern != m_pMemoryDDAPI->read32(currentAddress)))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 64:
                        if ((offset != tempOffset) &&
                            ((UINT64)in_u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress)))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                }

                if (bIsFailed)
                {
                    initProgress(0);

                    pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        strErrorMsg = "failed address: " + UINT64ToHexStr(rankOffset + tempOffset);
                        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                              strErrorMsg.c_str(),
                                              NULL);
                    }
                    else
                    {
                        return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                    }
                }

                if (tempOffset != 0)
                {
                    tempOffset <<= 1;
                }
                else
                {
                    // We should align the second offset address
                    tempOffset = u64DataByteWidth;
                }
            }

            // Check remaining address from rankOffset which will be power of 2.
            tempOffset = rankOffset << 1;
            while (tempOffset < lastRankOffset)
            {
                pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(tempOffset,
                                                               in_s32Controller,
                                                               in_s32Channel,
                                                               in_s32Rank,
                                                               &currentAddress);
                if (pErrorMsg != NO_ERROR_MSG)
                {
                    initProgress(0);
                    return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                }

                switch (in_s32ElementSize)
                {
                    case 8:
                        if ((UINT8)in_u64UserTestPattern != m_pMemoryDDAPI->read8(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 16:
                        if ((UINT16)in_u64UserTestPattern != m_pMemoryDDAPI->read16(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 32:
                        if ((UINT32)in_u64UserTestPattern != m_pMemoryDDAPI->read32(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 64:
                        if ((UINT64)in_u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                }

                if (bIsFailed)
                {
                    initProgress(0);

                    pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        strErrorMsg = "failed address: " + UINT64ToHexStr(tempOffset);
                        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                              strErrorMsg.c_str(),
                                              NULL);
                    }
                    else
                    {
                        return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                    }
                }

                tempOffset <<= 1;
            }

            // Write back test pattern
            pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(rankOffset + offset,
                                                           in_s32Controller,
                                                           in_s32Channel,
                                                           in_s32Rank,
                                                           &currentAddress);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)in_u64UserTestPattern);
                    break;
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)in_u64UserTestPattern);
                    break;
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)in_u64UserTestPattern);
                    break;
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT64)in_u64UserTestPattern);
                    break;
            }

            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            if (offset != 0)
            {
                offset <<= 1;
            }
            else
            {
                offset = u64DataByteWidth;
            }

            updateProgress(1);
        }

        // Verify from rankOffset which is power of 2
        offset = rankOffset << 1;
        while (offset < lastRankOffset)
        {
            // Write inverse test pattern
            pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(offset,
                                                           in_s32Controller,
                                                           in_s32Channel,
                                                           in_s32Rank,
                                                           &currentAddress);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)(~in_u64UserTestPattern));
                    break;
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)(~in_u64UserTestPattern));
                    break;
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)(~in_u64UserTestPattern));
                    break;
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT64)(~in_u64UserTestPattern));
                    break;
            }

            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            // Check remaining addresses from rankOffset plus an offset which will be power of 2.
            tempOffset = 0;
            while (tempOffset < testBlockSize)
            {
                pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress((rankOffset + tempOffset),
                                                               in_s32Controller,
                                                               in_s32Channel,
                                                               in_s32Rank,
                                                               &currentAddress);
                if (pErrorMsg != NO_ERROR_MSG)
                {
                    initProgress(0);
                    return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                }

                switch (in_s32ElementSize)
                {
                    case 8:
                        if ((UINT8)in_u64UserTestPattern != m_pMemoryDDAPI->read8(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 16:
                        if ((UINT16)in_u64UserTestPattern != m_pMemoryDDAPI->read16(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 32:
                        if ((UINT32)in_u64UserTestPattern != m_pMemoryDDAPI->read32(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 64:
                        if ((UINT64)in_u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                }

                if (bIsFailed)
                {
                    initProgress(0);

                    pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        strErrorMsg = "failed address: " + UINT64ToHexStr(rankOffset + tempOffset);
                        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                              strErrorMsg.c_str(),
                                              NULL);
                    }
                    else
                    {
                        return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                    }
                }

                if (tempOffset)
                {
                    tempOffset <<= 1;
                }
                else
                {
                    tempOffset = u64DataByteWidth;
                }
            }

            // Check remaining addresses from rankOffset which will be power of 2.
            tempOffset = rankOffset << 1;
            while (tempOffset < lastRankOffset)
            {
                pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(tempOffset,
                                                               in_s32Controller,
                                                               in_s32Channel,
                                                               in_s32Rank,
                                                               &currentAddress);
                if (pErrorMsg != NO_ERROR_MSG)
                {
                    initProgress(0);
                    return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                }

                switch (in_s32ElementSize)
                {
                    case 8:
                        if ((offset != tempOffset) &&
                            (UINT8)in_u64UserTestPattern != m_pMemoryDDAPI->read8(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 16:
                        if ((offset != tempOffset) &&
                            (UINT16)in_u64UserTestPattern != m_pMemoryDDAPI->read16(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 32:
                        if ((offset != tempOffset) &&
                            (UINT32)in_u64UserTestPattern != m_pMemoryDDAPI->read32(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 64:
                        if ((offset != tempOffset) &&
                            (UINT64)in_u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                }

                if (bIsFailed)
                {
                    initProgress(0);

                    pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        strErrorMsg = "failed address: " + UINT64ToHexStr(tempOffset);
                        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                              strErrorMsg.c_str(),
                                              NULL);
                    }
                    else
                    {
                        return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                    }
                }

                tempOffset <<= 1;
            }

            // Write back test pattern
            pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(offset,
                                                           in_s32Controller,
                                                           in_s32Channel,
                                                           in_s32Rank,
                                                           &currentAddress);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)in_u64UserTestPattern);
                    break;
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)in_u64UserTestPattern);
                    break;
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)in_u64UserTestPattern);
                    break;
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT64)in_u64UserTestPattern);
                    break;
            }

            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            offset <<= 1;

            updateProgress(1);
        }
    }
    else
    {
        offset = 0;
        while (offset < lastRankOffset)
        {
            // Write inverse test pattern
            pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(offset,
                                                           in_s32Controller,
                                                           in_s32Channel,
                                                           in_s32Rank,
                                                           &currentAddress);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)(~in_u64UserTestPattern));
                    break;
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)(~in_u64UserTestPattern));
                    break;
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)(~in_u64UserTestPattern));
                    break;
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT64)(~in_u64UserTestPattern));
                    break;
            }

            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            // Check remaining addresses from offset which will be power of 2.
            tempOffset = 0;
            while (tempOffset < testBlockSize)
            {
                pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(tempOffset,
                                                               in_s32Controller,
                                                               in_s32Channel,
                                                               in_s32Rank,
                                                               &currentAddress);
                if (pErrorMsg != NO_ERROR_MSG)
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                }

                switch (in_s32ElementSize)
                {
                    case 8:
                        if ((offset != tempOffset) &&
                            ((UINT8)in_u64UserTestPattern != m_pMemoryDDAPI->read8(currentAddress)))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 16:
                        if ((offset != tempOffset) &&
                            ((UINT16)in_u64UserTestPattern != m_pMemoryDDAPI->read16(currentAddress)))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 32:
                        if ((offset != tempOffset) &&
                            ((UINT32)in_u64UserTestPattern != m_pMemoryDDAPI->read32(currentAddress)))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                    case 64:
                        if ((offset != tempOffset) &&
                            ((UINT64)in_u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress)))
                        {
                            bIsFailed = TRUE;
                        }
                        break;
                }

                if (bIsFailed)
                {
                    initProgress(0);

                    pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        strErrorMsg = "failed address: " + UINT64ToHexStr(tempOffset);
                        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                              strErrorMsg.c_str(),
                                              NULL);
                    }
                    else
                    {
                        return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
                    }
                }

                if (tempOffset != 0)
                {
                    tempOffset <<= 1;
                }
                else
                {
                    // We should align the second offset address
                    tempOffset = u64DataByteWidth;
                }
            }

            // Write back test pattern
            pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(offset,
                                                           in_s32Controller,
                                                           in_s32Channel,
                                                           in_s32Rank,
                                                           &currentAddress);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)in_u64UserTestPattern);
                    break;
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)in_u64UserTestPattern);
                    break;
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)in_u64UserTestPattern);
                    break;
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT64)in_u64UserTestPattern);
                    break;
            }

            if (pErrorMsg != NO_ERROR_MSG)
            {
                initProgress(0);
                return returnErrorMsg(ERROR_MEMORY_DAPI_ADDRESS_LINE_TEST, NULL, pErrorMsg);
            }

            if (offset != 0)
            {
                offset <<= 1;
            }
            else
            {
                offset = u64DataByteWidth;
            }

            updateProgress(1);
        }
    }

    return NO_ERROR_MSG;
}
