/// @file       MemoryDAPI_ECCErrorDetectionTest.cpp
/// @brief      Provide ECC error detection test for MemoryDAPI
/// @details    Inject ECC error and check if it can be detected
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#include <list>
#include <string>

#include "MemoryDAPI.hpp"
#include "CommonFunctions.hpp"


/// @fn         MemoryDAPI::eccErrorDetectionTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                               PHYSICAL_ADDRESS_T in_endAddress)
/// @brief      Verify memory ECC pins
/// @param      in_startAddress is a physical address for start address
/// @param      in_endAddress is a physical address for end address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::eccErrorDetectionTest(PHYSICAL_ADDRESS_T in_startAddress,
                                  PHYSICAL_ADDRESS_T in_endAddress)
{
    ErrorMsg            *pErrorMsg = NO_ERROR_MSG, *pTempErrorMsg = NO_ERROR_MSG;
    PHYSICAL_ADDRESS_T  startAddress = in_startAddress;
    UINT64              u64Steps;
    
    startAddress = alignMemoryAddress(startAddress, 64);
    u64Steps = (in_endAddress - startAddress) >> 3;
    
    initProgress(u64Steps);
    
    while (startAddress < in_endAddress)
    {
        // Verify ECC[7:0] = 00000000b
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(startAddress, 0x0, 1)))
        {
            break;
        }
        // Verify ECC[0]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(startAddress, 0xC, 3)))
        {
            break;
        }
        // Verify ECC[1]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(startAddress, 0xA, 0)))
        {
            break;
        }
        // Verify ECC[2]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(startAddress, 0x9, 1)))
        {
            break;
        }
        // Verify ECC[3]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(startAddress, 0x41, 6)))
        {
            break;
        }
        // Verify ECC[4]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(startAddress, 0x4001, 13)))
        {
            break;
        }
        // Verify ECC[5]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(startAddress, 0x2001, 28)))
        {
            break;
        }
        // Verify ECC[6]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(startAddress, 0x10000001, 59)))
        {
            break;
        }

        updateProgress(1);

        startAddress += 8;
    }
    
    if (pErrorMsg != NO_ERROR_MSG)
    {
        initProgress(0);
        if ((pTempErrorMsg = m_pMemoryDDAPI->clearECCReport()) != NO_ERROR_MSG)
        {
            deleteErrorMsg(pTempErrorMsg);
        }
        return returnErrorMsg(ERROR_MEMORY_DAPI_ECC_ERROR_DETECTION_TEST, NULL, pErrorMsg);
    }
    else
    {
        if ((pErrorMsg = m_pMemoryDDAPI->clearECCReport()) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_MEMORY_DAPI_ECC_ERROR_DETECTION_TEST, NULL, pErrorMsg);
        }
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::eccErrorDetectionTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                               UINT64             in_u64TestSize,
///                                               INT32              in_s32Controller,
///                                               INT32              in_s32Channel,
///                                               INT32              in_s32Rank)
/// @brief      Verify memory ECC pins
/// @param      in_startAddress is a physical address for start address
/// @param      in_u64TestSize is an UINT64 value for test size
/// @param      in_s32Controller is an INT32 value for target controller
/// @param      in_s32Channel is an INT32 value for target channel
/// @param      in_s32Rank is an INT32 value for target rank
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::eccErrorDetectionTest(PHYSICAL_ADDRESS_T in_startAddress,
                                  UINT64             in_u64TestSize,
                                  INT32              in_s32Controller,
                                  INT32              in_s32Channel,
                                  INT32              in_s32Rank)
{
    ErrorMsg            *pErrorMsg = NO_ERROR_MSG, *pTempErrorMsg = NO_ERROR_MSG;
    PHYSICAL_ADDRESS_T  rankOffset = (PHYSICAL_ADDRESS_T)0;
    PHYSICAL_ADDRESS_T  lastRankOffset = 0;
    PHYSICAL_ADDRESS_T  currentAddress;
    UINT64              u64Steps;

    in_startAddress = alignMemoryAddress(in_startAddress, 64);

    // Get the rank offset based on the start address
    if ((pErrorMsg = m_pMemoryDDAPI->getRankOffset(in_startAddress, &rankOffset)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_ECC_ERROR_DETECTION_TEST, NULL, pErrorMsg);
    }

    in_u64TestSize >>= 3;
    u64Steps = in_u64TestSize;
    // Align for 8 bytes
    in_u64TestSize <<= 3;

    lastRankOffset = rankOffset + (PHYSICAL_ADDRESS_T)in_u64TestSize;

    initProgress(u64Steps);

    while (rankOffset < lastRankOffset)
    {
        // Get physical address for target rank with rank offset
        pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(rankOffset, in_s32Controller, in_s32Channel, in_s32Rank, &currentAddress);
        if (pErrorMsg != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_MEMORY_DAPI_ECC_ERROR_DETECTION_TEST, NULL, pErrorMsg);
        }

        // Verify ECC[7:0] = 00000000b
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(currentAddress, 0x0, 1)))
        {
            break;
        }
        // Verify ECC[0]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(currentAddress, 0xC, 3)))
        {
            break;
        }
        // Verify ECC[1]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(currentAddress, 0xA, 0)))
        {
            break;
        }
        // Verify ECC[2]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(currentAddress, 0x9, 1)))
        {
            break;
        }
        // Verify ECC[3]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(currentAddress, 0x41, 6)))
        {
            break;
        }
        // Verify ECC[4]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(currentAddress, 0x4001, 13)))
        {
            break;
        }
        // Verify ECC[5]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(currentAddress, 0x2001, 28)))
        {
            break;
        }
        // Verify ECC[6]
        if ((pErrorMsg = m_pMemoryDDAPI->injectECC(currentAddress, 0x10000001, 59)))
        {
            break;
        }

        updateProgress(1);

        rankOffset += 8;
    }

    if (pErrorMsg != NO_ERROR_MSG)
    {
        initProgress(0);
        if ((pTempErrorMsg = m_pMemoryDDAPI->clearECCReport()) != NO_ERROR_MSG)
        {
            deleteErrorMsg(pTempErrorMsg);
        }
        return returnErrorMsg(ERROR_MEMORY_DAPI_ECC_ERROR_DETECTION_TEST, NULL, pErrorMsg);
    }
    else
    {
        if ((pErrorMsg = m_pMemoryDDAPI->clearECCReport()) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_MEMORY_DAPI_ECC_ERROR_DETECTION_TEST, NULL, pErrorMsg);
        }
    }

    return NO_ERROR_MSG;
}
