/// @file       NVMeDAPI.cpp
/// @brief      NVMeDAPI class
/// @details    Provide diagnostic and utility APIs for NVMe module
/// @version    1.0.0
/// @date       2014.01.06
/// @author     Jerry Lee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "NVMeDAPI.hpp"
#include "MemoryDDAPI.hpp"
#include "CommonFunctions.hpp"
#include <string.h>
#include <iomanip>

/// @fn         NVMeDAPI::NVMeDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
/// @brief      NVMeDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
//NVMeDDAPI* NVMeDAPI::nvmeDDAPI = NULL;
NVMeDAPI::NVMeDAPI(STRING className) : BaseClassDAPI(className)
{
    // Initializations.
    //if (nvmeDDAPI == NULL)
    //{
    //    if((nvmeDDAPI = new NVMeDDAPI("NVMe_DRVR")) == NULL)
    //    {
    //        addErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_ERRORCODE | ERROR_DIAGNOSTIC_API_CONSTRUCTOR2_ERROR), "Create NVMe DDAPI fail");
    //    }
    //}
    
    dataBuffer = (UINT32*)NULL;
    logicalSectorToTest = 0;
    rwBlocks = 0;
    testSize = 0;
    transferSize = 0;
    throttle = 0;
    testPattern = 0;
    testSpeed = 0;
};

/// @fn         NVMeDAPI::~NVMeDAPI(void)
/// @brief      NVMeDAPI destructor
/// @exception  Nothing
/// @return     Nothing
NVMeDAPI::~NVMeDAPI(void)
{
    // delete NVMe Driver API
    //if(nvmeDDAPI != NULL)
    //    delete nvmeDDAPI;
};

/// @fn         activateFirmwareByController(UINT8 controllerID, nvme_fw_active active)
/// @brief      call driver ioctl to activate downloaded firmware
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::activateFirmwareByController(UINT8 controllerID, nvme_fw_active active)
{
    ErrorMsg* pErrorMsg = NULL;
    NVMeDDAPI nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_FW_ACTIVATE, NULL, pErrorMsg);
    }    
    
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::FIRMWARE_ACTIVATE, &active);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_FW_ACTIVATE, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         downloadFirmwareByController(UINT8 controllerID, const char *fw)
/// @brief      call driver ioctl to download firmware
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::downloadFirmwareByController(UINT8 controllerID, const char *fw)
{
    ErrorMsg*           pErrorMsg = NULL;
    FILE*               file = NULL;
    INT32               fileSize = 0;
    SYSTEM_ADDRESS_T    fwBuffer;
    MemoryDDAPI         memoryDDAPI("MemoryDDAPI");
    NVMeDDAPI           nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
        return returnErrorMsg(ERROR_NVME_DAPI_FW_DOWNLOAD, NULL, pErrorMsg); 
    
    if((pErrorMsg = memoryDDAPI.getErrorMsg()) != NO_ERROR_MSG)
        return returnErrorMsg(ERROR_NVME_DAPI_FW_DOWNLOAD, NULL, pErrorMsg);

    if ((file = fopen(fw, "r")) == NULL)
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_FW_DOWNLOAD | ERROR_OPEN_FILE), fw);
    // Allocate a memory buffer for FW
    fseek(file, 0, SEEK_END);
    fileSize = ftell(file);
    fseek(file, 0, SEEK_SET);

    if((pErrorMsg = memoryDDAPI.getMemoryBuffer(fileSize, &fwBuffer)) != NO_ERROR_MSG)
    {
        fclose(file);
        return returnErrorMsg(ERROR_NVME_DAPI_FW_DOWNLOAD, NULL, pErrorMsg);
    }

    if (fileSize != (INT32)fread((CHAR *)fwBuffer, 1, fileSize, file))
    {
        fclose(file);
        memoryDDAPI.releaseMemoryBuffer(fwBuffer);
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_FW_DOWNLOAD | ERROR_READ_FILE_FAIL), fw);
    }

    fclose(file);

    pErrorMsg = nvmeDDAPI.downloadFirmware(controllerID, (void *)fwBuffer, fileSize);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        memoryDDAPI.releaseMemoryBuffer(fwBuffer);
        return returnErrorMsg(ERROR_NVME_DAPI_FW_DOWNLOAD, NULL, pErrorMsg);
    }

    memoryDDAPI.releaseMemoryBuffer(fwBuffer);

    return NO_ERROR_MSG;
}

/// @fn         NVMeDAPI::doIOCTL
/// @brief      Execute specific IOCLT command
/// @param      Specific NVME controller
/// @param      NVMe ioctl operation code
/// @param      The returned value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::doIOCTL(UINT8 controllerID, UINT32 ioctlOperation, UINT32* returnedValue)
{
    ErrorMsg*   pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_DO_IOCTL, NULL, pErrorMsg);
    }
    
    // only support no argument input ioctl OP code
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, (NVMeIOCTL::IOCTL_OP_CODE)ioctlOperation, returnedValue);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_DO_IOCTL, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getBarResourceByController(UINT8 controllerID, nvme_bar **barInfo)
/// @brief      call driver ioctl to get controller's bar resource
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getBarResourceByController(UINT8 controllerID, nvme_bar **barInfo)
{
    int                 pci_fd;
    char                path[512];
    NVMe_DEVICE_INFO    nvmeInfo;
    string              tempName;
    ErrorMsg*           pErrorMsg = NULL;
    NVMeDDAPI           nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_BAR_RESOURCE, NULL, pErrorMsg);
    }
    
    if((pErrorMsg = getInformationByController(controllerID, &nvmeInfo)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_BAR_RESOURCE, NULL, pErrorMsg);
    }
    tempName = nvmeInfo.nativeName;
    transform(tempName.begin(), tempName.end(), tempName.begin(), ::tolower);
    sprintf(path, "/sys/class/nvme/%s/device/resource0", tempName.c_str());

    pci_fd = open(path, O_RDONLY);
    if (pci_fd < 0)
    {
        sprintf(path, "/sys/class/misc/%s/device/resource0", nvmeInfo.chardeviceNodeName.c_str());
        pci_fd = open(path, O_RDONLY);
    }
    if (pci_fd < 0)
        return returnErrorMsg(ERROR_NVME_DAPI_GET_BAR_RESOURCE, "can not find pci resource", NULL);

    *barInfo = (nvme_bar *)mmap(0, getpagesize(), PROT_READ, MAP_SHARED, pci_fd, 0);
    if (*barInfo == MAP_FAILED)
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_GET_BAR_RESOURCE | ERROR_NVME_DDAPI_MMAP_FAIL), NULL, NULL);

    return NO_ERROR_MSG;
}

/// @fn         getErrorLogByController(UINT8 controllerID, nvme_error_log_page errLog)
/// @brief      call driver ioctl to get controller's error log
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getErrorLogByController(UINT8 controllerID, nvme_error_log_page* errLog)
{
    ErrorMsg*   pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_ERROR_LOG, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::ERROR_LOG, errLog);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_ERROR_LOG, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getFirmwareLogByController(UINT8 controllerID, nvme_firmware_log_page* fwLog)
/// @brief      call driver ioctl to get controller's firmware log
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getFirmwareLogByController(UINT8 controllerID, nvme_firmware_log_page* fwLog)
{
    ErrorMsg*   pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_FW_LOG, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::FIRMWARE_LOG, fwLog);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_FW_LOG, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         NVMeDAPI::getInformationByController(UINT8 controllerID)
/// @brief      call driver ioctl to get NVMe controller basic informations
/// @param      NVME ID
/// @param      pointer to return NVMe Info structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getInformationByController(UINT8 controllerID, NVMeInfo* returnInfo)
{
    ErrorMsg*    pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.getControllerList(&nvmeList);
    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    if(nvmeList.empty())
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }
    if(controllerID > nvmeList.size())
    {
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_GET_INFORMATION | ERROR_BAD_PARAMETER),
                              "Bad NVME logical ID");
    }
    for(iter = nvmeList.begin(); iter != nvmeList.end(); iter++)
    {
        if (controllerID == (iter - nvmeList.begin()))
        {
            *returnInfo = *iter;
            return NO_ERROR_MSG;
        }
    }

    return returnErrorMsg(ERROR_NVME_DAPI_GET_INFORMATION, NULL, pErrorMsg);
}

/// @fn         NVMeDAPI::getInformationList(vector<NVMe_DEVICE_INFO>* returnedList)
/// @brief      Get NVMe information list
/// @param      pointer to NVMe information list on the host
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getInformationList(vector<NVMe_DEVICE_INFO>* returnList)
{
    ErrorMsg*   pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.getControllerList(returnList);

    if(returnList->empty())
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getIdentityCTRLByController(UINT8 controllerID, nvme_id_ctrl* idCtrl)
/// @brief      call driver ioctl to get controller's information
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getIdentityCTRLByController(UINT8 controllerID, nvme_id_ctrl* idCtrl)
{
    ErrorMsg*   pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_IDENTITY_CONTROLLER, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::CONTROLLER_INFORMATION, idCtrl);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_IDENTITY_CONTROLLER, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getIdentityNSByController(UINT8 controllerID, nvme_id_ns* idNS)
/// @brief      call driver ioctl to get controller's namespace information
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getIdentityNSByController(UINT8 controllerID, nvme_id_ns* idNS)
{
    ErrorMsg*    pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_IDENTITY_NAMESPACE, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::NAMESPACE_INFORMATION, idNS);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_IDENTITY_NAMESPACE, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         NVMeDAPI::getLogicalIDByName(string wellKnowName, UINT8* returnID)
/// @brief      Get NVMe controller logical ID
/// @param      Controller well known name
/// @param      pointer to NVMe controller logical ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getLogicalIDByName(string wellKnowName, UINT8* returnID)
{
    ErrorMsg*   pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_LOGICALID_BY_NAME, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.getControllerList(&nvmeList);

    if(nvmeList.empty())
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_LOGICALID_BY_NAME, NULL, pErrorMsg);
    }

    ConvertStringToUpperCase((char*)wellKnowName.c_str());

    for(iter = nvmeList.begin(); iter != nvmeList.end(); iter++)
    {
        if (wellKnowName == iter->wellKnowName)
        {
            *returnID = (iter - nvmeList.begin());
            return NO_ERROR_MSG;
        }
    }

    return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_GET_LOGICALID_BY_NAME | ERROR_BAD_PARAMETER_FIVE),
                          NULL, NULL);
}

/// @fn         getNameSpaceIDByController(UINT8 controllerID, int* returnValue)
/// @brief      call driver ioctl to get namespace ID
/// @param      NVME ID
/// @param      pointer to return ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getNameSpaceIDByController(UINT8 controllerID, int* returnValue)
{
    ErrorMsg*   pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_NAMESPACE_ID, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::NAMESPACE_ID, (void *)returnValue);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_NAMESPACE_ID, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getSmartLogByController(UINT8 controllerID, nvme_smart_log* smartLog)
/// @brief      call driver ioctl to get smart log
/// @param      NVME ID
/// @param      pointer to return smart log structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getSmartLogByController(UINT8 controllerID, nvme_smart_log* smartLog)
{
    ErrorMsg*   pErrorMsg = NULL;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_SMART_LOG, NULL, pErrorMsg);
    }
    
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::SMART_LOG, smartLog);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_SMART_LOG, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         NVMeDAPI::resetCtrl(UINT8 controllerID)
/// @brief      reset specific NVME controller
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::resetCtrl(UINT8 controllerID)
{
	ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_RESET_CTRL, NULL, pErrorMsg);
    }
    
	pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::EXECUTE_NVME_RESET, NULL);

    if(pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_RESET_CTRL, NULL, pErrorMsg);
    }
    
    return NO_ERROR_MSG;
}

/// @fn 		NVMeDAPI::resetSubSystem(UINT8 controllerID)
/// @brief		reset specific NVME Subsystem
/// @param		NVME ID
/// @exception	Nothing
/// @return 	A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::resetSubSystem(UINT8 controllerID)
{
	ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_RESET_SUBSYS, NULL, pErrorMsg);
    }
    
	//pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::EXECUTE_RESET_NVME_SUBSYS, NULL);
	pErrorMsg = nvmeDDAPI.resetSubSystem(controllerID);
	
	if(pErrorMsg != NO_ERROR_MSG)
	{
		return returnErrorMsg(ERROR_NVME_DAPI_RESET_SUBSYS, NULL, pErrorMsg);
	}
	
	return NO_ERROR_MSG;
}


ErrorMsg*
NVMeDAPI::readFromDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize)
{
    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

ErrorMsg*
NVMeDAPI::writeToDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize)
{
    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

/// @fn         NVMeDAPI::writeZeroByController(UINT8 controllerID, UINT64 startBlock, UINT16 blockCount);
/// @brief      Write zero to specified NVMe block
/// @param      Specific NVME controller
/// @param      start block to write
/// @param      block count to write zero
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::writeZeroByController(UINT8 controllerID, UINT64 startBlock, UINT16 blockCount)
{
    ErrorMsg*               pErrorMsg = NULL;
    nvme_write_zero_parms   parms;
    NVMeDDAPI               nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_WRITE_ZERO, NULL, pErrorMsg);
    }

    parms.startBlock = startBlock;
    parms.blockCount = blockCount;

    if((pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::NAMESPACE_ID, (void *)&parms.nsid)) == NO_ERROR_MSG)
    {
        pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::WRITE_ZERO, &parms);
    }

    if(pErrorMsg != NO_ERROR_MSG)
        return returnErrorMsg(ERROR_NVME_DAPI_WRITE_ZERO, NULL, pErrorMsg);

    return NO_ERROR_MSG;
}


ErrorMsg*
NVMeDAPI::getFeatures(UINT8 controllerID, nvme_get_features* getFeatures)
{
    ErrorMsg*       pErrorMsg = NULL;
	CHAR            tempString[(DEBUG_MESSAGE_LENGTH + 1)];
    UINT8           *u8buf;
	UINT8           field;
	UINT64          ull;
	UINT32          i, j;
    UINT32          incSize;
	UINT32          offDuration;
    MemoryDDAPI     memoryDDAPI("MemoryDDAPI");
    NVMeDDAPI       nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);
    }
    
	if (getFeatures->sel > 3)
	{
        sprintf(tempString, "\"Select\" parameter %d is invalid!\n", getFeatures->sel);
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_GET_FEATURES | ERROR_BAD_PARAMETER_ONE), tempString, NULL);

	}
	if (getFeatures->fid == 0) {
        sprintf(tempString, "\"Feature ID\" parameter %d is invalid!\n", getFeatures->fid);
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_GET_FEATURES | ERROR_BAD_PARAMETER_TWO), tempString, NULL);
	}

	if(getFeatures->fid == NVME_FEAT_LBA_RANGE)
	{
		getFeatures->data_len = 4096;
	}
	else if (getFeatures->fid == NVME_FEAT_AUTO_PST)
	{
		getFeatures->data_len = 256;
	}
	else if (getFeatures->fid == NVME_FEAT_HOST_ID)
	{
		getFeatures->data_len = 8;
	}
    else if ((getFeatures->fid == NVME_FEAT_MAX_LBA) || (getFeatures->fid == NVME_FEAT_NATIVE_MAX_LBA))
	{
		getFeatures->data_len = 8;
	}   

	if (getFeatures->data_len != 0)
	{	
	    if((pErrorMsg = memoryDDAPI.getErrorMsg()) != NO_ERROR_MSG)
	        return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);

	    // Allocate a memory buffer
	    if((pErrorMsg = memoryDDAPI.getMemoryBuffer(getFeatures->data_len, (SYSTEM_ADDRESS_T*)&getFeatures->data_buf)) != NO_ERROR_MSG)
	    {
	        return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);
	    }
        memset(getFeatures->data_buf, 0, getFeatures->data_len);
/*
//printf("getFeatures->data_len: %d\n", getFeatures->data_len);
printf("getFeatures->result: %d\n", getFeatures->result);
//printf("*getFeatures->data_buf : 0x%llX\n", *(UINT64*)getFeatures->data_buf);
printf("*getFeatures->data_buf : 0x%08X\n", *(UINT32*)getFeatures->data_buf);
//printf("*(getFeatures->data_buf + 4) : 0x%08X\n", *(UINT32*)(getFeatures->data_buf + 4));
*/


	}

    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::GET_FEATURES, (void*)getFeatures);

    if(pErrorMsg != NO_ERROR_MSG)
    {
    	if (getFeatures->data_buf != NULL)
	        memoryDDAPI.releaseMemoryBuffer((SYSTEM_ADDRESS_T)getFeatures->data_buf);
		
        return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);
    }

	switch (getFeatures->fid)
	{
		case NVME_FEAT_ARBITRATION:
        
            printf("Arbitration\n");
            printf("---------------------------------------------------------------\n");
			printf("High Priority Weight   : %u\n", ((getFeatures->result & 0xFF000000) >> 24) + 1);
			printf("Medium Priority Weight : %u\n", ((getFeatures->result & 0xFF0000) >> 16) + 1);
			printf("Low Priority Weight    : %u\n", ((getFeatures->result & 0xFF00) >> 8) + 1);
			printf("Arbitration Burst      : %u\n",  (1 << (getFeatures->result & 0x7)));
			break;

		case NVME_FEAT_POWER_MGMT:
            printf("Power Management\n");
            printf("---------------------------------------------------------------\n");
			field = (getFeatures->result & 0xE0) >> 5;
			printf("Workload Hint : ");
            switch (field)
            {
                case 0:	
                    printf("No Workload\n");
                    break;
                case 1:
                    printf("Extended Idle Period with a Burst of Random Writes\n");
                    break;
                case 2:
                    printf("Heavy Sequential Writes\n");
                    break;
                default:
                    printf("Reserved\n");
                    break;
            }
			printf("Power State   : %u\n",  getFeatures->result & 0x1F);
			break;

		case NVME_FEAT_LBA_RANGE:
            printf("LBA Range Type\n");
            printf("---------------------------------------------------------------\n");
			field = getFeatures->result & 0x3F;
			printf("Number of LBA Ranges      : %u\n", field + 1);
            
            struct nvme_lba_range_type *lbrt;
            lbrt = (struct nvme_lba_range_type *)getFeatures->data_buf;
            for (i = 0; i <= field; i++)
            {
                printf("Type                      : ");
                switch (lbrt[i].type)
                {
                    case 0:	
                        printf("Reserved\n");
                        break;
                    case 1:
                        printf("Filesystem\n");
                        break;
                    case 2:
                        printf("RAID\n");
                        break;
                    case 3:
                        printf("Cache\n");
                        break;
                    case 4:
                        printf("Page / Swap file\n");
                        break;
                    default:
                        if ((lbrt[i].type >= 0x5 ) && (lbrt[i].type <= 0x7F))
                        {
                            printf("Reserved\n");
                        }
                        else
                        {
                            printf("Vendor Specific\n");
                        }
                        
                        break;
                }
    
                printf("Attributes                : %s, %s\n", (lbrt[i].attributes & 0x0001) ? "LBA range may be overwritten":"LBA range should not be overwritten",
                    ((lbrt[i].attributes & 0x2) >> 1) ? "LBA range should be hidden from the OS/EFI/BIOS":"LBA range should be visible from the OS/EFI/BIOS");
                printf("Starting LBA              : %#llX\n", (UINT64)(lbrt[i].slba));
                printf("Number of Logical Blocks  : %#llX\n", (UINT64)(lbrt[i].nlb));
                printf("Unique Identifier (GUID)  : ");
                for (j = 0; j < 16; j++)
                    printf("%02X", lbrt[i].guid[j]);
                printf("\n");
            }
			break;

		case NVME_FEAT_TEMP_THRESH:
            printf("Temperature Threshold\n");
            printf("---------------------------------------------------------------\n");
			field = (getFeatures->result & 0x300000) >> 20;
			printf("Threshold Type Select        : ");
            switch (field)
            {
                case 0:
                    printf("Over Temperature Threshold\n");
                    break;
                case 1:
                    printf("Under Temperature Threshold\n");
                    break;
                default:
                    printf("Reserved\n");
                    break;
            }

			field = (getFeatures->result & 0xF0000) >> 16;
			printf("Threshold Temperature Select : ");
            switch (field)
            {
                case 0:
                    printf("Composite Temperature\n");
                    break;
                case 1:
                    printf("Temperature Sensor 1\n");
                    break;
                case 2:
                    printf("Temperature Sensor 2\n");
                    break;
                case 3:
                    printf("Temperature Sensor 3\n");
                    break;
                case 4:
                    printf("Temperature Sensor 4\n");
                    break;
                case 5:
                    printf("Temperature Sensor 5\n");
                    break;
                case 6:
                    printf("Temperature Sensor 6\n");
                    break;
                case 7:
                    printf("Temperature Sensor 7\n");
                    break;
                case 8:
                    printf("Temperature Sensor 8\n");
                    break;
                default:
                    printf("Reserved\n");
                    break;
            }
           
			printf("Temperature Threshold        : %u C\n", (getFeatures->result & 0xFFFF) - 273);
			break;
		case NVME_FEAT_ERR_RECOVERY:
            printf("Error Recovery\n");
            printf("---------------------------------------------------------------\n");
			printf("Deallocated or Unwritten Logical Block Error Enable : %s\n", ((getFeatures->result & 0x10000) >> 16) ? "Enable":"Disable");
			printf("Time Limited Error Recovery                         : %u ms\n", (getFeatures->result & 0xFFFF) * 100);
			break;
		case NVME_FEAT_VOLATILE_WC:
            printf("Volatile Write Cache\n");
            printf("---------------------------------------------------------------\n");
			printf("Volatile Write Cache Enable : %s\n", (getFeatures->result & 0x1) ? "Enable":"Disable");
			break;
		case NVME_FEAT_NUM_QUEUES:
            printf("Number of Queues\n");
            printf("---------------------------------------------------------------\n");
			printf("Number of IO Completion Queues Allocated : %u\n", ((getFeatures->result & 0xFFFF0000) >> 16) + 1);
			printf("Number of IO Submission Queues Allocated : %u\n",  (getFeatures->result & 0xFFFF) + 1);
			break;
		case NVME_FEAT_IRQ_COALESCE:
            printf("Interrupt Coalescing\n");
            printf("---------------------------------------------------------------\n");
			printf("Aggregation Time      : %u ms\n", ((getFeatures->result & 0xFF00) >> 8) * 100);
			printf("Aggregation Threshold : %u\n",  (getFeatures->result & 0xFF) + 1);
			break;
		case NVME_FEAT_IRQ_CONFIG:
            printf("Interrupt Vector Configuration\n");
            printf("---------------------------------------------------------------\n");
			printf("Coalescing Disable : %s\n", ((getFeatures->result & 0x10000) >> 16) ? "True":"False");
			printf("Interrupt Vector   : %u\n",  getFeatures->result & 0xFFFF);
			break;
		case NVME_FEAT_WRITE_ATOMIC:
            printf("Write Atomicity Normal\n");
            printf("---------------------------------------------------------------\n");
			printf("Disable Normal : %s\n", (getFeatures->result & 0x1) ? "True":"False");
			break;
		case NVME_FEAT_ASYNC_EVENT:
            printf("Asynchronous Event Configuration\n");
            printf("---------------------------------------------------------------\n");
			printf("Firmware Activation Notices      : %s\n", ((getFeatures->result & 0x200) >> 9) ? "Send async event":"Do not send async event");
			printf("Namespace Attribute Notices      : %s\n", ((getFeatures->result & 0x100) >> 8) ? "Send NameSpace Attribute Changed event":"Do not send NameSpace Attribute Changed event");
			printf("SMART / Health Critical Warnings : %s\n", (getFeatures->result & 0xFF) ? "Send async event":"Do not send async event");
			break;
		case NVME_FEAT_AUTO_PST:
            struct nvme_auto_pst * autoPst;
            autoPst = (struct nvme_auto_pst *)getFeatures->data_buf;

            printf("Autonomous Power State Transition\n");
            printf("---------------------------------------------------------------\n");
			printf("Autonomous Power State Transition Enable : %s\n", (getFeatures->result & 0x1) ? "Enable":"Disable");
            for (i = 0; i < 32; i++)
            {
                printf("[Entry %02d] Idle Time Prior to Transition : %u ms\n", i, (autoPst[i].data & 0xFFFFFF00) >> 8);
            }
            for (i = 0; i < 32; i++)
            {
                printf("[Entry %02d] Idle Transition Power State   : %u\n",  i, (autoPst[i].data & 0xF8) >> 3);
            }
			break;
		case NVME_FEAT_HOST_MEM_BUF:
            struct nvme_host_mem_buffer *hostMemBuffer;
            hostMemBuffer = (struct nvme_host_mem_buffer *)getFeatures->data_buf;
            printf("Host Memory Buffer\n");
            printf("---------------------------------------------------------------\n");
			printf("Memory Return                           : %s\n", ((getFeatures->result & 0x2) >> 1) ? "True":"False");
			printf("Enable Host Memory                      : %s\n", (getFeatures->result & 0x1) ? "Enable":"Disable");
            printf("Host Memory Descriptor List Entry Count : %u\n", hostMemBuffer->hmdlec);
            printf("Host Memory Descriptor List Address     : %u\n", hostMemBuffer->hmdlau);
            printf("Host Memory Descriptor List Address     : %u\n", hostMemBuffer->hmdlal);
            printf("Host Memory Buffer Size                 : %u\n", hostMemBuffer->hsize);
			break;
		case NVME_FEAT_KEEP_ALIVE_TIMER:
            printf("Keep Alive Timer\n");
            printf("---------------------------------------------------------------\n");
			printf("Keep Alive Timeout : ");
            if (getFeatures->result)
                printf("%u ms\n", getFeatures->result);
            else
                printf("Disabled\n");
			break;
            
		case NVME_FEAT_SW_PROGRESS:
            printf("Software Progress Marker\n");
            printf("---------------------------------------------------------------\n");
			printf("Pre-boot Software Load Count : %u\n", getFeatures->result & 0xFF);
			break;
		case NVME_FEAT_HOST_ID:
            printf("Host Identifier\n");
            printf("---------------------------------------------------------------\n");            
            printf("Enable Extended Host Identifier : %s\n", (getFeatures->result & 0x1) ? "Yes":"NO");
            u8buf = (UINT8*)getFeatures->data_buf;
            ull = ((UINT64)u8buf[7] << 56) | ((UINT64)u8buf[6] << 48) | ((UINT64)u8buf[5] << 40) | ((UINT64)u8buf[4] << 32);
            ull |= ((UINT64)u8buf[3] << 24) | ((UINT64)u8buf[2] << 16) | ((UINT64)u8buf[1] << 8) | u8buf[0];
            printf("Host Identifier (HOSTID):  %llX \n", ull);
			break;
		case NVME_FEAT_RESV_MASK:
            printf("Reservation Notification Mask\n");
            printf("---------------------------------------------------------------\n");
			printf("Mask Reservation Preempted Notification  : %s\n", ((getFeatures->result & 0x8) >> 3) ? "True":"False");
			printf("Mask Reservation Released Notification   : %s\n", ((getFeatures->result & 0x4) >> 2) ? "True":"False");
			printf("Mask Registration Preempted Notification : %s\n", ((getFeatures->result & 0x2) >> 1) ? "True":"False");
			break;
		case NVME_FEAT_RESV_PERSIST:
            printf("Reservation Persistance\n");
            printf("---------------------------------------------------------------\n");
			printf("Persist Through Power Loss : %s\n", (getFeatures->result & 0x1) ? "True":"False");
			break;
		case NVME_FEAT_ENDRC_MANAG_DISABLE:
            printf("Get Endurance Management Disable Setting\n");
            printf("---------------------------------------------------------------\n");
			printf("Endurance Management Disabled Setting: %s\n", (getFeatures->result & 0x1) ? "Disable":"Enable");
			break;
		case NVME_FEAT_MAX_LBA:
            printf("Get Max LBA\n");
            printf("---------------------------------------------------------------\n");
            printf("Max LBA : %#llX\n", *(UINT64*)getFeatures->data_buf);
            
			break;
		case NVME_FEAT_NATIVE_MAX_LBA:
            printf("Get Native Max LBA\n");
            printf("---------------------------------------------------------------\n");
            printf("Max LBA : %#llX\n", *(UINT64*)getFeatures->data_buf);
			break;
		case NVME_FEAT_POWER_GOVERNOR:
            printf("Get Power Governor Setting\n");
            printf("---------------------------------------------------------------\n");
			field = getFeatures->result & 0xFF;
			printf("Power Governor Setting: ");
			if (field == 0x0)
				printf("25W\n");
			else if (field == 0x1)
				printf("20W\n");
			else if (field == 0x2)
				printf("15W\n");
			else
				printf("Unknown\n");
			break;
		case NVME_FEAT_SMB_ASIC_ADDR:
            printf("Get SMB ASIC Address\n");
            printf("---------------------------------------------------------------\n");
			field = (getFeatures->result & 0x3FE) >> 1;
			printf("SMB Controller Address: %#X", field);
			if (field != 0xFF)
				printf("\n");
			else
				printf(" (FW disables SMB ASIC access)\n");
			break;
		case NVME_FEAT_BLINK_ACTLED:
            printf("Get Blink Activity for LED\n");
            printf("---------------------------------------------------------------\n");

            //Feature Option 0
            printf("LED state while host is inactive:    ");
            printf("%s\n", getFeatures->result ? "ON":"OFF");
			
			//Feature Option 1
            getFeatures->cdw11 = 0x01000000;
            pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::GET_FEATURES, (void*)getFeatures);
            if(pErrorMsg != NO_ERROR_MSG)               
                return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);
            //printf("LED duration increment size: ");
            //printf("%s\n", getFeatures->result ? "25ms":"50ms");
            if (getFeatures->result == 0)
                incSize = 50;
            else
                incSize = 25;
                
			//Feature Option 2
            getFeatures->cdw11 = 0x02000000;
            pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::GET_FEATURES, (void*)getFeatures);
            if(pErrorMsg != NO_ERROR_MSG)               
                return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);
            //printf("OFF duration during IO activity in 25ms/50ms increments: ");
            printf("OFF duration during IO activity:     ");
            offDuration = getFeatures->result;
            if (offDuration == 0)
            {
                printf("Solid ON\n");
                
            }
            else
            {
                offDuration *= incSize;
                printf("%u ms\n", offDuration);
            }

			//Feature Option 3
            getFeatures->cdw11 = 0x03000000;
            pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::GET_FEATURES, (void*)getFeatures);
            if(pErrorMsg != NO_ERROR_MSG)               
                return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);
			//printf("ON duration during IO activity in 25ms/50ms increments: ");
            printf("ON duration during IO activity:      ");
            if (getFeatures->result == 0)
            {
                //printf("Match OFF duration\n");
                if (offDuration == 0)
                    printf("Solid ON\n");
                else
                    printf("%u ms\n", offDuration);
            }
            else
                printf("%u ms\n", (getFeatures->result * incSize));

			//Feature Option 4
            getFeatures->cdw11 = 0x04000000;
            pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::GET_FEATURES, (void*)getFeatures);
            if(pErrorMsg != NO_ERROR_MSG)               
                return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);
            //printf("OFF duration during format activity in 25ms/50ms increments: ");
            printf("OFF duration during format activity: ");
            offDuration = getFeatures->result;
            if (offDuration == 0)
            {
                printf("Solid ON\n");
                
            }
            else
            {
                offDuration *= incSize;
                printf("%u ms\n", offDuration);
            }
            
			//Feature Option 5
            getFeatures->cdw11 = 0x05000000;
            pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::GET_FEATURES, (void*)getFeatures);
            if(pErrorMsg != NO_ERROR_MSG)               
                return returnErrorMsg(ERROR_NVME_DAPI_GET_FEATURES, NULL, pErrorMsg);
            //printf("ON duration during format activity in 25ms/50ms increments: ");
            printf("ON duration during format activity:  ");
            if (getFeatures->result == 0)
            {
                //printf("Match OFF duration\n");
                if (offDuration == 0)
                    printf("Solid ON\n");
                else
                    printf("%u ms\n", offDuration);
            }
            else
                printf("%u ms\n", (getFeatures->result * incSize));


			break;
		case NVME_FEAT_ENABLE_LATENCY_TRACK:
            printf("Get Enable Latency Tracking\n");
            printf("---------------------------------------------------------------\n");
            printf("Latency Tracking: ");
			if (getFeatures->result & 0x1)
				printf("Enable\n");
			else
				printf("Disable\n");
                
			break;
        default:
            printf("?\n");
            break;
	}

	if (getFeatures->data_buf != NULL)
		memoryDDAPI.releaseMemoryBuffer((SYSTEM_ADDRESS_T)getFeatures->data_buf);

    return NO_ERROR_MSG;
}

ErrorMsg*
NVMeDAPI::setFeatures(UINT8 controllerID, nvme_set_features* setFeatures)
{
    ErrorMsg*	pErrorMsg = NULL;
	CHAR		tempString[(DEBUG_MESSAGE_LENGTH + 1)];
    NVMeDDAPI   nvmeDDAPI("NVMeDDAPI");
        
    if ((pErrorMsg = nvmeDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVME_DAPI_SET_FEATURES, NULL, pErrorMsg);
    }
    
	if (setFeatures->fid == 0) {
        sprintf(tempString, "Invalid 'feature id' parameter %d!\n", setFeatures->fid);
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_SET_FEATURES | ERROR_BAD_PARAMETER_ONE), tempString, NULL);
	}
/*
	if (setFeatures.feature_id == NVME_FEAT_LBA_RANGE)
		setFeatures->data_len = 4096;

	if (setFeatures->data_len != 0)
	{	
	    if((pErrorMsg = memoryDDAPI.getErrorMsg()) != NO_ERROR_MSG)
	        return returnErrorMsg(ERROR_NVME_DAPI_SET_FEATURES, NULL, pErrorMsg);

	    // Allocate a memory buffer
	    if((pErrorMsg = memoryDDAPI.getMemoryBuffer(setFeatures->data_len, (SYSTEM_ADDRESS_T*)setFeatures->data_buf)) != NO_ERROR_MSG)
	    {
	        return returnErrorMsg(ERROR_NVME_DAPI_SET_FEATURES, NULL, pErrorMsg);
	    }
	}
*/
	
    pErrorMsg = nvmeDDAPI.doIOCTLCommand(controllerID, NVMeIOCTL::SET_FEATURES, (void*)setFeatures);

    if(pErrorMsg != NO_ERROR_MSG)
    {
    	//if (setFeatures->data_buf != NULL)
	    //    memoryDDAPI.releaseMemoryBuffer((SYSTEM_ADDRESS_T)setFeatures->data_buf);
        return returnErrorMsg(ERROR_NVME_DAPI_SET_FEATURES, NULL, pErrorMsg);
    }

	//if (setFeatures->data_buf != NULL)
	//	memoryDDAPI.releaseMemoryBuffer((SYSTEM_ADDRESS_T)setFeatures->data_buf);

    return NO_ERROR_MSG;
}

void
NVMeDAPI::displayRaw(unsigned char *buf, int len)
{
	int     i;
    bool    finalLine = false;
	char    ascii[17];
    
	for (i = 0; i < len; i++)
	{
		finalLine = 0;
        
		if (i % 16 == 0)
			printf("%04X:", i);
		
		if (i % 8 == 0)
			printf("  ");

		printf("%02X ", buf[i]);
        
        if ((buf[i] >= 0x20) && (buf[i] <= 0x7E))
            ascii[i % 16] = buf[i];
        else
            ascii[i % 16] = '.';
            
        if ((i + 1) % 16 == 0)
        {
            ascii[i % 16 + 1] = '\0';
            printf(" |  %s\n", ascii);
            finalLine = true;
        }
	}

	if (!finalLine)
	{
        if (i % 16 <= 8)
            printf("%*s", 50 - (i % 16) * 3, "");
        else
            printf("%*s", 48 - (i % 16) * 3, "");
            
        ascii[i % 16] = '\0';
        printf(" |  %s\n", ascii);
	}

}
