/// @file       NVMeDAPI.hpp
/// @brief      The header file for NVMe DAPI class
/// @details    Provide diagnostic and utility APIs for NVMe module
/// @version    1.0.0
/// @date       2014.01.06
/// @author     Jerry Lee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#ifndef NVMeDAPI_H
#define NVMeDAPI_H

#include <vector>
#include <sys/mman.h>
#include "BaseClassDAPI.hpp"
#include "NVMeDDAPI.hpp"

// NVMe_DATA_ELEMENT definitions
typedef enum {
    NVMe_DATA_IOC_NUM = 0,
    NVMe_DATA_IOC_DRIVER,
    NVMe_DATA_IOC_ID,
    NVMe_DATA_DEVICE_TYPE,
    NVMe_DATA_BACK_CHNL_TRNG_STATUS,
    NVMe_DATA_INVALID_DWORD_ERR_COUNT,
    NVMe_DATA_DISPARITY_ERR_COUNT,
    NVMe_DATA_LOSS_DWORD_SYNC_COUNT,
    NVMe_DATA_PHY_RESET_PRBM_COUNT,
    NVMe_DATA_ENABLE_PHY,
    NVMe_DATA_DISABLE_PHY,
    NVMe_DATA_ERROR_STATISTICS,
    NVMe_DATA_JITTER_ERROR_COUNT,
    NVMe_DATA_LINK_STATUS,
    NVMe_DATA_LINK_SPEED,
    NVMe_DATA_IOC_TEMPERATURE
} NVMe_DATA_ELEMENT;

// Performance information definitions
typedef struct NVME_PERFORMANCE_INFORMATION {
        string DeviceNode;
        UINT64 StartLogicalSector;
        UINT64 TestSize;
        timeval RunTimes;
        double m_dSpeed;
} NVMePerformanceInfo;

/// @class      NVMeDAPI NVMeDAPI.hpp "NVMeDAPI.hpp"
/// @brief      Provide diagnostic and utility APIs for NVMe module
/// @param      m_pNVMeDDAPI is a pointer to NVMeDDAPI
class NVMeDAPI: public BaseClassDAPI
{
    public:

                NVMeDAPI(STRING className);

        virtual ~NVMeDAPI(void);

        ErrorMsg* activateFirmwareByController(UINT8 controllerID, nvme_fw_active active);
        ErrorMsg* doIOCTL(UINT8 controllerID, UINT32 ioctlOperation, UINT32* returnValue);
        ErrorMsg* downloadFirmwareByController(UINT8 controllerID, const char *fw);
        ErrorMsg* getBarResourceByController(UINT8 controllerID, nvme_bar **barInfo);
        ErrorMsg* getErrorLogByController(UINT8 controllerID, nvme_error_log_page* errLog);
        ErrorMsg* getFirmwareLogByController(UINT8 controllerID, nvme_firmware_log_page* fwLog);
        ErrorMsg* getInformationByController(UINT8 controllerID, NVMeInfo* returnInfo);
        ErrorMsg* getInformationList(vector<NVMe_DEVICE_INFO>* returnList);
        ErrorMsg* getLogicalIDByName(string wellKnowName, UINT8* returnID);
        ErrorMsg* getSmartLogByController(UINT8 controllerID, nvme_smart_log* smartLog);
        ErrorMsg* getIdentityCTRLByController(UINT8 controllerID, nvme_id_ctrl* idCtrl);
        ErrorMsg* getIdentityNSByController(UINT8 controllerID, nvme_id_ns* idNS);
        ErrorMsg* getNameSpaceIDByController(UINT8 controllerID, int* returnValue);
		ErrorMsg* resetCtrl(UINT8 controllerID);
		ErrorMsg* resetSubSystem(UINT8 controllerID);
		ErrorMsg* readFromDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize);
		ErrorMsg* writeToDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize);
		ErrorMsg* writeZeroByController(UINT8 controllerID, UINT64 startBlock, UINT16 blockCount);
		ErrorMsg* getFeatures(UINT8 controllerID, nvme_get_features* getFeatures);
		ErrorMsg* setFeatures(UINT8 controllerID, nvme_set_features* setFeatures);

		// test main entry
		ErrorMsg* readPerformanceTest(NVMePerformanceInfo* performanceInformation);
		ErrorMsg* writePerformanceTest(NVMePerformanceInfo* performanceInformation);
		ErrorMsg* readWriteTest(NVMePerformanceInfo* performanceInformation);

    protected:

    private:
		//static NVMeDDAPI* nvmeDDAPI;
		vector<NVMe_DEVICE_INFO> nvmeList;
		vector<NVMe_DEVICE_INFO>::iterator iter;

		struct timeval spendTime;
		string  deviceNode;
		UINT32* dataBuffer;
		UINT64  logicalSectorToTest;
		UINT32  rwBlocks;
		UINT64  testSize;
		UINT32  transferSize;
		UINT32  throttle;
		UINT32  testPattern;
		double  testSpeed;

		ErrorMsg* readPerformanceTestSetup(void);
		ErrorMsg* readPerformanceTestExecute(void);
		ErrorMsg* readPerformanceTestCleanup(void);

		ErrorMsg* readWriteTestSetup(void);
		ErrorMsg* readWriteTestExecute(void);
		ErrorMsg* readWriteTestCleanup(void);

		ErrorMsg* writePerformanceTestSetup(void);
		ErrorMsg* writePerformanceTestExecute(void);
		ErrorMsg* writePerformanceTestCleanup(void);

		void displayRaw(unsigned char *buf, int len);
};

#endif  // NVMeDAPI_H
