/// @file       MemoryDAPI_AddressLineTest.cpp
/// @brief      Provide address line test for MemoryDAPI
/// @details    Verify memory address lines by write and read test pattern for power of 2 addressing
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "CommonFunctions.hpp"
#include "NVSRAMDAPI.hpp"


/// @fn         NVSRAMDAPI::nonVolatileTest(PHYSICAL_ADDRESS_T startAddress, PHYSICAL_ADDRESS_T endAddress, INT32 elementSize)
/// @brief      NVSRAM non volatile RAM test
/// @param      startAddress is a physcial address for start address
/// @param      endAddress is a physcial address for end address
/// @param      ElementSize is 32bits value for element size (ex: 8/16/32/64)
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVSRAMDAPI::nonVolatileTest
    (
    PHYSICAL_ADDRESS_T startAddress,
    PHYSICAL_ADDRESS_T endAddress,
    INT32              elementSize
    )
{
    ErrorMsg* ErrorMsg = NULL;
    PHYSICAL_ADDRESS_T  addressTemp;
    UINT32 requestedSize, index = 0;
    UINT8 data = 0;

    if((ErrorMsg = getInformationList(&testNvsramList)) != NO_ERROR_MSG)
    {
        return ErrorMsg;
    }

    testIter = testNvsramList.begin();

    // validate the user input
    if ((elementSize != 8) && (elementSize != 16) && (elementSize != 32) && (elementSize != 64))
    {

        return returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST | ERROR_ILLEGAL_ELEMENT_SIZE),
                                          "Unknown element size",
                                          NULL);
    }
    if ((startAddress > endAddress) || (endAddress > testIter->BARSize))
    {
        return returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST | ERROR_OUT_OF_RANGE),
                                          "Out of NVSRAM range",
                                          NULL);
    }
    // backup SRAM content into backup buffer
    requestedSize = endAddress - startAddress;
    ErrorMsg = memDAPI->getMemoryBuffer(requestedSize, &bufferBackup);
    ErrorMsg = memDAPI->getMemoryBuffer(requestedSize, &nvsramBackup);

    for (addressTemp = startAddress+testIter->BaseAddress; addressTemp <= endAddress+testIter->BaseAddress; addressTemp++)
    {
        data = memDAPI->readMemory8(addressTemp);
        *((UINT8 *)bufferBackup + index) = data;
        index++;
    }
    // backup non-volatile memory content into backup buffer
    sleep(3);
    // issue RECALL command to
    nvsramDDAPI->recallNVSRAM();
    sleep(3);

    index = 0;
    for (addressTemp = startAddress+testIter->BaseAddress; addressTemp <= endAddress+testIter->BaseAddress; addressTemp++)
    {
        data = memDAPI->readMemory8(addressTemp);
        *((UINT8 *)nvsramBackup + index) = data;
        index++;
    }
    if (ErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVSRAM_DAPI_NVRAM_TEST, NULL, ErrorMsg);
    }

    return NO_ERROR_MSG;
}


/// @fn         NVSRAMDAPI::NVTestExecute(PHYSICAL_ADDRESS_T startAddress, PHYSICAL_ADDRESS_T endAddress, INT32 elementSize)
/// @brief      NVSRAM non volatile RAM test
/// @param      startAddress is a physcial address for start address
/// @param      endAddress is a physcial address for end address
/// @param      ElementSize is 32bits value for element size (ex: 8/16/32/64)
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVSRAMDAPI::NVTestExecute
    (
    PHYSICAL_ADDRESS_T startAddress,
    PHYSICAL_ADDRESS_T endAddress,
    INT32              elementSize
    )
{
    ErrorMsg* ErrorMsg = NULL;
    UINT16 patternCount;
    UINT16 testNumber;
    UINT8 testPatternByte;
    UINT64 testPattern;
    // validate the user input
    if ((elementSize != 8) && (elementSize != 16) && (elementSize != 32) && (elementSize != 64))
    {
        return returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST | ERROR_ILLEGAL_ELEMENT_SIZE),
                                         "Unknown element size",
                                          NULL);
    }
    patternCount = elementSize / 8;
    // execute the test for each test pattern
    for(testNumber = 0; testNumber < 4; testNumber++)
    {
        switch(testNumber)
        {
            case 0:
                    testPatternByte = DFW_API_TEST_PATTERN_BYTE_ONE;
                    break;
            case 1:
                    testPatternByte = DFW_API_TEST_PATTERN_BYTE_TWO;
                    break;
            case 2:
                    testPatternByte = DFW_API_TEST_PATTERN_BYTE_THREE;
                    break;
            default:
                    testPatternByte = DFW_API_TEST_PATTERN_BYTE_FOUR;
                    break;
        }
        // expand the test pattern to the requested element size
        testPattern = testPatternByte;
        if (patternCount >= 2)
            testPattern = (testPattern << 8) | testPatternByte;
        if (patternCount >= 4)
            testPattern = (testPattern << 16) | testPattern;
        if (patternCount >= 6)
            testPattern = (testPattern << 32) | testPattern;

        // call the actual test
        if ((ErrorMsg = NVTestExecute(startAddress, endAddress, elementSize, testPattern)) != NO_ERROR_MSG)
            break;
    }
    // return the error code
    if (ErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVSRAM_DAPI_NVRAM_TEST, NULL, ErrorMsg);
    }

    return NO_ERROR_MSG;
}


/* *************************************************************************
 * NVSRAMNonVolatileTesExecute
 *
 * Input:   startAddress    the address to start testing
 *          endAddress      the final address to test
 *          elementSize     the size of element to test (8/16/32)
 *          userTestPattern the user provided test pattern
 *
 * Returns: ERRORCODE
 *
 * ************************************************************************* */
ErrorMsg*
NVSRAMDAPI::NVTestExecute
    (
    PHYSICAL_ADDRESS_T startAddress,
    PHYSICAL_ADDRESS_T endAddress,
    INT32              elementSize,
    UINT64             userTestPattern
    )
{
    ErrorMsg* ErrorMsg = NULL;

    // validate the user input
    if ((elementSize != 8) && (elementSize != 16) && (elementSize != 32) && (elementSize != 64))
    {
        return returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST | ERROR_ILLEGAL_ELEMENT_SIZE),
                                          "Unknown element size",
                                          NULL);
    }
    // step 1 Write the test pattern into the device's SRAM
    // Initializations
    testStartAddress = testIter->BaseAddress + startAddress;
    testEndAddress = testIter->BaseAddress + endAddress;

    // Write patterns to memory
    if ((ErrorMsg = WritePattern(elementSize, userTestPattern)) != NO_ERROR_MSG)
    {
        returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST), NULL, ErrorMsg);
    }

    // step 2
    sleep(3);
    nvsramDDAPI->storeNVSRAM();//STORE command
    sleep(3);

    // step 3 Clear the NVSRAM "SAM memory" and verify it contains the cleared data
    // Clear memory
    if ((ErrorMsg = WritePattern(elementSize, ~userTestPattern)) != NO_ERROR_MSG)
    {
        returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST), NULL, ErrorMsg);
    }

    //Verify the NVSRAM SRAM contains the clear data
    if ((ErrorMsg = VerifyPattern(elementSize,~userTestPattern)) != NO_ERROR_MSG)
    {
        returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST), NULL, ErrorMsg);
    }

    // step 4
    sleep(3);
    nvsramDDAPI->recallNVSRAM();//RECALL command
    sleep(3);

    // step 5 Verify the NVSRAM SRAM contains the test pattern
    if ((ErrorMsg = VerifyPattern(elementSize,userTestPattern)) != NO_ERROR_MSG)
    {
        returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST), NULL, ErrorMsg);
    }
    // return the error code
    if (ErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVSRAM_DAPI_NVRAM_TEST, NULL, ErrorMsg);
    }

    return NO_ERROR_MSG;
}


/* *************************************************************************
 * NVSRAMNonVolatileTestCleanup
 *
 * Input:   startAddress    the address to start testing
 *          endAddress      the final address to test
 *          elementSize     the size of element to test (8/16/32)
 *
 * Returns: ERRORCODE
 *
 * ************************************************************************* */
ErrorMsg*
NVSRAMDAPI::NVTestCleanup
    (
    PHYSICAL_ADDRESS_T startAddress,
    PHYSICAL_ADDRESS_T endAddress,
    INT32              elementSize
    )
{
    ErrorMsg* ErrorMsg = NULL;
    PHYSICAL_ADDRESS_T addressTemp;
    UINT32 index = 0;

    // validate the user input
    if ((elementSize != 8) && (elementSize != 16) && (elementSize != 32) && (elementSize !=64))
    {
        return returnErrorMsg((ERRORCODE)(ERROR_NVSRAM_DAPI_NVRAM_TEST | ERROR_ILLEGAL_ELEMENT_SIZE),
                              "Unknown element size",
                              NULL);
    }
    // restore NVRAM content
    if (nvsramBackup != (SYSTEM_ADDRESS_T)NULL)
    {
        for (addressTemp = startAddress+testIter->BaseAddress; addressTemp <= endAddress+testIter->BaseAddress; addressTemp++)
        {
            memDAPI->writeMemory8(addressTemp, (UINT8)(*((UINT8 *)nvsramBackup + index)));
            index++;
        }
    }
    sleep(3);
    nvsramDDAPI->storeNVSRAM();
    sleep(3);
    // restore SRAM content
    index = 0;
    if (bufferBackup != (PHYSICAL_ADDRESS_T)NULL)
    {
        for (addressTemp = startAddress+testIter->BaseAddress; addressTemp <= endAddress+testIter->BaseAddress; addressTemp++)
        {
            memDAPI->writeMemory8(addressTemp, (UINT8)(*((UINT8 *)bufferBackup+index)));
            index++;
        }
    }
    // return the error code
    if (ErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_NVSRAM_DAPI_NVRAM_TEST, NULL, ErrorMsg);
    }

    return NO_ERROR_MSG;
}


/* ************************************************************************* */
/*                              Private Functions                            */
/* ************************************************************************* */
ErrorMsg*
NVSRAMDAPI::WritePattern(INT32 elementSize, UINT64 userTestPattern)
{
    ErrorMsg* ErrorMsg = NULL;
    UINT8 value8;
    UINT16 value16;
    UINT32 value32;
    UINT64 value64;
    PHYSICAL_ADDRESS_T testAddress;
    // Initializations
    value8 = userTestPattern;
    value16 = userTestPattern;
    value32 = userTestPattern;
    value64 = userTestPattern;
    // Write patterns to memory
    testAddress = testStartAddress;

    while ((testAddress <= testEndAddress) && (ErrorMsg == NO_ERROR_MSG))
    {
        // write the memory locations
        switch(elementSize)
        {
            case 8:
                memDAPI->writeMemory8(testAddress, value8);
                testAddress++;
                memDAPI->writeMemory8(testAddress, value8);
                testAddress++;
                memDAPI->writeMemory8(testAddress, value8);
                testAddress++;
                memDAPI->writeMemory8(testAddress, value8);
                testAddress++;
                break;
            case 16:
                memDAPI->writeMemory16(testAddress, value16);
                testAddress++;
                testAddress++;
                memDAPI->writeMemory16(testAddress, value16);
                testAddress++;
                testAddress++;
                break;
            case 32:
                memDAPI->writeMemory32(testAddress, value32);
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                break;
            case 64:
                memDAPI->writeMemory64(testAddress, value64);
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                break;
            default:
                return returnErrorMsg((ERRORCODE)ERROR_ILLEGAL_ELEMENT_SIZE, NULL, NULL);
        }
    }

    return NO_ERROR_MSG;
}


ErrorMsg*
NVSRAMDAPI::VerifyPattern(INT32 elementSize, UINT64 userTestPattern)
{
    ErrorMsg* ErrorMsg = NULL;
    UINT8 value8;
    UINT16 value16;
    UINT32 value32;
    UINT64 value64;
    PHYSICAL_ADDRESS_T testAddress;
    // Initializations
    value8 = userTestPattern;
    value16 = userTestPattern;
    value32 = userTestPattern;
    value64 = userTestPattern;
    // Read patterns from memory
    testAddress = testStartAddress;

    while ((testAddress <= testEndAddress) && (ErrorMsg == NO_ERROR_MSG))
    {
        // read/verify the memory locations
        switch(elementSize)
        {
            case 8:
                if (memDAPI->readMemory8(testAddress) != value8)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                testAddress++;
                if (memDAPI->readMemory8(testAddress) != value8)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                testAddress++;
                if (memDAPI->readMemory8(testAddress) != value8)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                testAddress++;
                if (memDAPI->readMemory8(testAddress) != value8)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                testAddress++;
                break;
            case 16:
                if (memDAPI->readMemory16(testAddress) != value16)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                testAddress++;
                testAddress++;
                if (memDAPI->readMemory16(testAddress) != value16)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                testAddress++;
                testAddress++;
                break;
            case 32:
                if (memDAPI->readMemory32(testAddress) != value32)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                break;
            case 64:
                if (memDAPI->readMemory64(testAddress) != value64)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                testAddress++;
                break;
            default:
                return returnErrorMsg((ERRORCODE)ERROR_ILLEGAL_ELEMENT_SIZE, NULL, NULL);
        }
    }

    return NO_ERROR_MSG;
}
