/// @file       SASDAPI.hpp
/// @brief      The header file for SAS DAPI class
/// @details    Provide diagnostic and utility APIs for SAS module
/// @version    1.0.0
/// @date       2014.01.06
/// @author     Jerry Lee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#ifndef SASDAPI_H
#define SASDAPI_H

#include <vector>
#include "BaseClassDAPI.hpp"
#include "SASDDAPI.hpp"
#include "SCSIFactory.hpp"
#include "MemoryDDAPI.hpp"
//using std::vector;

#define SAS_RW_SIZE (1024*1024)
#define MEGABYTES (1024*1024)
#define DEFAULT_SAS_TEST_SIZE 0x1400000   // 20MB

// define the supported SAS loopback paths
#define SAS_LOOPBACK_NONE               0
#define SAS_LOOPBACK_EXTERNAL           1
#define SAS_LOOPBACK_INTERNAL           2
#define SAS_LOOPBACK_WITH_EXPANDER      3
#define SAS_LOOPBACK_REVERSE_EXTERNAL 	4

// SAS_DATA_ELEMENT definitions
typedef enum {
    SAS_DATA_IOC_NUM = 0,
    SAS_DATA_IOC_DRIVER,
    SAS_DATA_IOC_ID,
    SAS_DATA_DEVICE_TYPE,
    SAS_DATA_BACK_CHNL_TRNG_STATUS,
    SAS_DATA_INVALID_DWORD_ERR_COUNT,
    SAS_DATA_DISPARITY_ERR_COUNT,
    SAS_DATA_LOSS_DWORD_SYNC_COUNT,
    SAS_DATA_PHY_RESET_PRBM_COUNT,
    SAS_DATA_ENABLE_PHY,
    SAS_DATA_DISABLE_PHY,
    SAS_DATA_ERROR_STATISTICS,
    SAS_DATA_JITTER_ERROR_COUNT,
    SAS_DATA_LINK_STATUS,
    SAS_DATA_LINK_SPEED,
    SAS_DATA_IOC_TEMPERATURE
} SAS_DATA_ELEMENT;

// Performance information definitions
typedef struct PERFORMANCE_INFORMATION {
        string DeviceNode;
        UINT64 StartLogicalSector;
        UINT64 TestSize;
        timeval RunTimes;
        double m_dSpeed;
} PerformanceInfo;

/// @class      SASDAPI SASDAPI.hpp "SASDAPI.hpp"
/// @brief      Provide diagnostic and utility APIs for SAS module
/// @param      m_pSASDDAPI is a pointer to SASDDAPI
class SASDAPI: public BaseClassDAPI
{
    public:

                SASDAPI(STRING className);

        virtual ~SASDAPI(void);

        ErrorMsg* clearPHYErrorCountByController(UINT8 controllerID);
        ErrorMsg* clearPHYErrorCountByPHY(UINT8 controllerID, UINT8 phyNumber);
        ErrorMsg* doIOCTL(UINT8 controllerID, UINT32 ioctlOperation, UINT32* returnValue);
        ErrorMsg* doLoopbackTest(UINT8 controllerID);
        ErrorMsg* getAttachDeviceByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnType);
        ErrorMsg* getAddressByController(UINT8 controllerID, UINT64* returnAddress);
        ErrorMsg* getDriveDeviceNodeListByController(UINT8 controllerID, vector<SAS_DRIVE_INFO>* nodeList);
        ErrorMsg* getErrorCountByPHY(UINT8 controllerID, UINT8 phyNumber, ErrorCountInfo* returnInfo);
        ErrorMsg* getErrorCountByController(UINT8 controllerID, ErrorCountInfo* returnInfo);
        ErrorMsg* getInformationByController(UINT8 controllerID, SasInfo* returnInfo);
        ErrorMsg* getInformationByDriveName(string wellKnownName, SasDriveInfo* returnInfo);
        ErrorMsg* getInformationList(vector<SAS_CONTROLLER_INFO>* returnList);
        ErrorMsg* getLinkSpeedByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnSpeed);
        ErrorMsg* getLinkStatusByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnStatus);
        ErrorMsg* getLogicalIDByName(string wellKnowName, UINT8* returnID);
        ErrorMsg* getSmartDataByDriveName(string wellKnownName, StorageSASSMARTDataStruct* smartData);
        ErrorMsg* getTemperatureByController(UINT8 controllerID, UINT32* returnTemperature);
        ErrorMsg* programSASAddress(UINT8 controllerID, string sasAddress);
        ErrorMsg* programSASFirmware(UINT8 controllerID, string filepath);
		ErrorMsg* resetIOC(UINT8 controllerID);
		ErrorMsg* readFromDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize);
		ErrorMsg* writeToDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize);

    protected:

    private:
		SASDDAPI* sasDDAPI;
		MemoryDDAPI* memDDAPI;
		SCSIFactory* scsiFactory;

		struct timeval spendTime;
		string  deviceNode;
		UINT32* dataBuffer;
		UINT64  logicalSectorToTest;
        UINT32  rwBlocks;
		UINT64  testSize;
		UINT32  transferSize;
		UINT32  throttle;
		UINT32  testPattern;
		double  testSpeed;
};

#endif  // SASDAPI_H
