/// @file       PCIDev.cpp
/// @brief      PCIDev class
/// @details    Provide basic information and operation for a PCI device
/// @author     KenHu
/// @copyright  Copyright 2017 Hon Hai Technology Group. All Rights Reserved.

#include "PCIDev.hpp"

PCIDev::PCIDev()
{
    m_u16Domain = (UINT16)UNKNOWN_PCI_VALUE;
    m_u8Bus = (UINT8)UNKNOWN_PCI_VALUE;
    m_u8Dev = (UINT8)UNKNOWN_PCI_VALUE;
    m_u8Func = (UINT8)UNKNOWN_PCI_VALUE;
    initializeInfo();
}

PCIDev::PCIDev(UINT16 in_u16Domain, UINT8 in_u8Bus, UINT8 in_u8Dev, UINT8 in_u8Func)
{
    m_u16Domain = in_u16Domain;
    m_u8Bus = in_u8Bus;
    m_u8Dev = in_u8Dev;
    m_u8Func = in_u8Func;
    initializeInfo();
}

PCIDev::~PCIDev()
{
    m_capList.clear();
}

UINT64
PCIDev::getBAR64(INT32 in_s32BarID)
{
    INT32 i = 0;
    list<UINT32> barList;
    UINT64 u64Addr = 0ULL;

    list<UINT32>::iterator it;

    barList = getBARList();

    for (it = barList.begin(); it != barList.end(); it++, i++)
    {
        if (i == in_s32BarID)
        {
            u64Addr = (UINT64)*it;
            u64Addr |= ((UINT64)*(++it)) << 32;
            return u64Addr;
        }
        else
        {
            it++;
        }
    }

    return u64Addr;
}

list<UINT32>
PCIDev::getBARList()
{
    UINT8   u8HeaderType;
    UINT32  u32BarOffset, u32EndOffset;

    if (m_barList.empty())
    {
        u32BarOffset = PCI_BAR_START_OFFSET;

        u8HeaderType = ConfigReadByte(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, 0xe);
        if((u8HeaderType & 0x1) == 0x1)
        {
            u32EndOffset = PCI_BAR_START_OFFSET + 0x4; // 2 BAR
        }
        else
        {
            u32EndOffset = PCI_BAR_START_OFFSET + 5 * 0x4; // 6 BAR
        }

        for (; u32BarOffset <= u32EndOffset; u32BarOffset += 0x4)
        {
            m_barList.push_back(ConfigReadDword(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, u32BarOffset));
        }
    }

    return m_barList;
}


UINT32
PCIDev::getCapabilityOffset(UINT32  in_u32CapID)
{
    list<CapInfo>::iterator it;

    initializeCapList();

    for(it = m_capList.begin(); it != m_capList.end(); ++it)
    {
        if(it->GetCapID() == in_u32CapID && it->GetCapOffset() < 0x100)
        {
            return it->GetCapOffset();
        }
    }

    return (UINT32)UNKNOWN_PCI_VALUE;
}

UINT16
PCIDev::getDeviceID()
{
    if (m_u16DeviceID == (UINT16)UNKNOWN_PCI_VALUE)
    {
        m_u16DeviceID = ConfigReadWord(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, 0x2);
    }
    return m_u16DeviceID;
}

UINT8
PCIDev::getPortTypeRawData()
{
    UINT32  u32PCIeCapOffset;
    UINT16  u16Temp;
    UINT8   u8PortType;

    u32PCIeCapOffset = getCapabilityOffset(PCIE_CAPABILITY_ID);
    if (u32PCIeCapOffset != UNKNOWN_PCI_VALUE)
    {
        u16Temp = ConfigReadWord(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, (u32PCIeCapOffset + PCIE_CAPABILITIES_REG_OFFSET));
        u8PortType = (u16Temp & 0xf0) >> 4;
    }
    else
    {
        u8PortType = 0xff;
    }

    return u8PortType;
}

UINT16
PCIDev::getVendorID()
{
    if (m_u16VendorID == (UINT16)UNKNOWN_PCI_VALUE)
    {
        m_u16VendorID = ConfigReadWord(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, 0x0);
    }
    return m_u16VendorID;
}

void
PCIDev::initializeCapList()
{
    UINT32  u32Temp;
    UINT32  u32TempCapID;
    UINT32  u32TempCapOffset;
    UINT32  u32TempNextCapOffset;
    UINT8   u8CapPtrValue;

    if (!m_capList.empty())
    {
        return;
    }

    u8CapPtrValue = ConfigReadWord(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, 0x34);

    if (u8CapPtrValue != 0x0)
    {
        u32Temp = ConfigReadDword(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, (UINT32)u8CapPtrValue);
        u32TempCapID = u32Temp & 0xff;
        u32TempNextCapOffset = (u32Temp & 0xff00) >> 8;

        while (u32TempNextCapOffset != 0x0)
        {
            CapInfo tempCapInfo(u32TempCapID, u32TempCapOffset, u32TempNextCapOffset);
            u32TempCapOffset = u32TempNextCapOffset;
            m_capList.push_back(tempCapInfo);
            u32Temp = ConfigReadDword(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, u32TempNextCapOffset);
            u32TempCapID = u32Temp & 0xff;
            u32TempNextCapOffset = (u32Temp & 0xff00) >> 8;
        }

        CapInfo tempCapInfo(u32TempCapID, u32TempCapOffset, u32TempNextCapOffset);
        m_capList.push_back(tempCapInfo);
    }

    initializeExtendedCapList();
}

void
PCIDev::initializeExtendedCapList()
{
    UINT32    u32Temp;
    UINT32    u32TempCapID;
    UINT32    u32TempCapOffset;
    UINT32    u32TempNextCapOffset;

    u32Temp = ConfigReadDword(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, 0x100);
    u32TempCapID = u32Temp & 0xffff;
    u32TempNextCapOffset = (u32Temp & 0xfff00000) >> 20;

    while (u32TempNextCapOffset != 0x0)
    {
        CapInfo tempCapInfo(u32TempCapID, u32TempCapOffset, u32TempNextCapOffset);
        u32TempCapOffset = u32TempNextCapOffset;
        m_capList.push_back(tempCapInfo);
        u32Temp = ConfigReadDword(m_u16Domain, m_u8Bus, m_u8Dev, m_u8Func, u32TempNextCapOffset);
        u32TempCapID = u32Temp & 0xffff;
        u32TempNextCapOffset = (u32Temp & 0xfff00000) >> 20;
    }

    CapInfo tempCapInfo(u32TempCapID, u32TempCapOffset, u32TempNextCapOffset);
    m_capList.push_back(tempCapInfo);
}

void
PCIDev::initializeInfo()
{
    m_pLowerBus = NULL;
    m_u16VendorID = (UINT16)UNKNOWN_PCI_VALUE;
    m_u16DeviceID = (UINT16)UNKNOWN_PCI_VALUE;
    m_u32Class = UNKNOWN_PCI_VALUE;
}

