/// @file       SASDDAPI.cpp
/// @brief      SASDDAPI class
/// @details    Provide low layer diagnostic and utility APIs for SAS module
/// @version    1.0.0
/// @date       2014.02.21
/// @author     Jerry Lee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.
#include <string.h>
#include <cstdlib>
#include <algorithm>
#include "SASDDAPI.hpp"
#define MAX_PHYS_OF_PORT   4
#define MPT2SAS_FILE_PATH  "/dev/mpt3ctl"
#define NUMOFHARDDRIVE     24
#define LSI_VENDOR_ID      0x1000
#define PMC_VENDOR_ID      0x11F8


/// @fn         SASDDAPI::SASDDAPI(STRING className) : BaseClassDDAPI(className)
/// @brief      SASDDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
SASDDAPI::SASDDAPI(STRING className) : BaseClassDDAPI(className)
{
    // initialize here
    if((pciDDAPI = new PCIDDAPI("PCI_DRVR")) == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_ERRORCODE | ERROR_DIAGNOSTIC_DRIVER_API_CONSTRUCTOR1_ERROR), "Create PCI DDAPI fail");
    }
    if((sasIOCTL = new SASIOCTL(MPT2SAS_FILE_PATH)) == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_ERRORCODE | ERROR_DIAGNOSTIC_DRIVER_API_CONSTRUCTOR1_ERROR), "Create SAS IOCTL fail");
    }
    if((scsiFactory = new SCSIFactory("SCSIFactory")) == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_ERRORCODE | ERROR_DIAGNOSTIC_DRIVER_API_CONSTRUCTOR1_ERROR), "Create SCSI Factory fail");
    }

    addIOCController(); // search and add IOC information on the host
}

/// @fn         SASDDAPI::~SASDDAPI(void)
/// @brief      SASDDAPI destructor
/// @exception  Nothing
/// @return     Nothing
SASDDAPI::~SASDDAPI(void)
{
    // delete PCI Driver API
    if (pciDDAPI != NULL)
        delete pciDDAPI;

    // delete IOCTL object we created
    if (sasIOCTL != NULL)
        delete sasIOCTL;

    // delete IOCTL object we created
    if (scsiFactory != NULL)
        delete scsiFactory;
}

/// @fn         SASDAPI::addIOCController(void)
/// @brief      add all IOCs' information on the host
/// @param      None
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDDAPI::addIOCController(void)
{
    mpt3_ioctl_iocinfo mptIOCInfo;
    SAS_CONTROLLER_INFO sasInfo;
    vector<PCIInformantion> pciInfo;
    PFA bdf;

    m_errorMsg = pciDDAPI->getPciInfoByClass(PCI_SAS, &pciInfo);
    if(m_errorMsg)
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_CONSTRUCTOR), NULL, m_errorMsg);

    for(vector<PCIInformantion>::iterator iter = pciInfo.begin(); iter != pciInfo.end(); iter++)
    {
        //BDF
        bdf.u16Domain = iter->u16Domain;
        bdf.u16Bus = iter->u16Bus;
        bdf.u16Device = iter->u16Device;
        bdf.u8Function = iter->u8Function;
        //PCI link speed
        sasInfo.pciSpeed = pciDDAPI->GetFloatDataByID(bdf, PCI_DATA_TYPE_CURRENT_SPEED);
        //PCI width
        sasInfo.pciWidth = pciDDAPI->GetDataByBDF(bdf, PCI_DATA_TYPE_CURRENT_WIDTH);
        switch (iter->u16VendorID)
        {
            case LSI_VENDOR_ID:
                if(pciInfo.size() == 2)
                {
                    sasInfo.wellKnowName = "SAS" + Int2Str(sasControllerIDwithNoHBA[iter - pciInfo.begin()]);
                    sasIOCTL->setControllerID((UINT8)(furyRealID[sasControllerIDwithNoHBA[iter - pciInfo.begin()]]));
                }
                else
                {
                    sasInfo.wellKnowName = "SAS" + Int2Str(sasControllerID[iter - pciInfo.begin()]);
                    sasIOCTL->setControllerID((UINT8)(furyRealID[sasControllerID[iter - pciInfo.begin()]]));
                }

                m_errorMsg = doIOCTLCommand(SASIOCTL::IOC_INFORMATION, &mptIOCInfo);

                if(m_errorMsg == (ErrorMsg *) NO_ERROR)
                {
                    sasInfo.driverVersion.assign((const char*) &mptIOCInfo.driver_version);
                    sasInfo.firmwareVersion = Int2Str((mptIOCInfo.firmware_version & 0xFF000000) >> 24) + "."
                            + Int2Str((mptIOCInfo.firmware_version & 0x00FF0000) >> 16) + "."
                            + Int2Str((mptIOCInfo.firmware_version & 0x0000FF00) >> 8) + "."
                            + Int2Str(mptIOCInfo.firmware_version & 0x000000FF);
                    sasInfo.biosVersion = Int2Str((mptIOCInfo.bios_version & 0xFF000000) >> 24) + "."
                                        + Int2Str((mptIOCInfo.bios_version & 0x00FF0000) >> 16) + "."
                                        + Int2Str((mptIOCInfo.bios_version & 0x0000FF00) >> 8) + "."
                                        + Int2Str(mptIOCInfo.bios_version & 0x000000FF);
                    sasInfo.domindBusDeviceFunction = Int2HexStr(iter->u16Domain,4) + ":"
                                                    + Int2HexStr(iter->u16Bus) + ":"
                                                    + Int2HexStr(iter->u16Device) + "."
                                                    + Int2HexStr(iter->u8Function, 1);
                    sasInfo.hardwareRevision = "Rev " + Int2HexStr(mptIOCInfo.hw_rev);

                    sasList.push_back(sasInfo);
                }

                break;
            case PMC_VENDOR_ID:

                break;
        }
    }

    std::sort(sasList.begin(), sasList.end(), sortSAScontroller);

    if(m_errorMsg)
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_CONSTRUCTOR), NULL, m_errorMsg);
    else
        return NO_ERROR_MSG;
}

/// @fn         SASDDAPI::doIOCTLCommand
/// @brief      execute SAS controller IOCTL command
/// @param      SAS_IOCTL_REQUEST
/// @param      The returned value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDDAPI::doIOCTLCommand(SASIOCTL::IOCTL_OP_CODE ioctlCommand, void* returnedValue)
{
    if((m_errorCode = sasIOCTL->allocateIOCTLArgument(ioctlCommand)) == NO_ERROR)
    {
        if ((m_errorCode = sasIOCTL->setIOCTLOperationArgument(ioctlCommand)) == NO_ERROR)
        // separate ioctl command here to distinguish the command has return data or customized argument
        switch(ioctlCommand)
        {
            case SASIOCTL::CLEAR_ERR_LOG:
            case SASIOCTL::CLEAR_ERR_LOG_BY_PHY:
            case SASIOCTL::EXECUTE_LOOPBACK:
            case SASIOCTL::EXECUTE_IOC_SMP:
            case SASIOCTL::EXECUTE_IOC_HARD_RESET:
                m_errorCode = sasIOCTL->executeIOCTL(ioctlCommand);

                break;
            case SASIOCTL::CURRENT_TEMP:
            case SASIOCTL::IOC_INFORMATION:
            case SASIOCTL::IOC_SAS_ADDRESS:
                m_errorCode = sasIOCTL->executeIOCTL(ioctlCommand, returnedValue);

                if(sasIOCTL->getPHYNumber() != INVALID_PHY_NUMBER)
                    m_errorCode = ERROR_SAS_DDAPI_NOT_SUPPORT_SPECIFIC_PHY;

                break;
            case SASIOCTL::SHOW_ATTACHED_DEVICE:
            case SASIOCTL::PHY_SAS_ADDRESS:
            case SASIOCTL::ERROR_STATISTICS:
            case SASIOCTL::PHY_ERROR_COUNT:
            case SASIOCTL::LINK_STATUS:
            case SASIOCTL::LINK_SPEED:
                m_errorCode = sasIOCTL->executeIOCTL(ioctlCommand, returnedValue);

                break;
            case SASIOCTL::UnspportIOCTLCommand:
                m_errorCode = ERROR_IOCTL_INVALID_COMMAND_CODE;

                break;
        }
    }

    if(m_errorCode != NO_ERROR)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_EXECUTE_IOCTL | m_errorCode), NULL, NULL);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDDAPI::getDeviceList
/// @brief      Get SAS information list on the host
/// @param      IOC ID
/// @param      image buffer
/// @param      image buffer size
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDDAPI::getControllerList(vector<SAS_CONTROLLER_INFO>* returnedList)
{
    if(sasList.empty())
        return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_SAS_LIST | ERROR_SAS_DDAPI_NO_SAS_CONTROLLER), NULL, NULL);

    *returnedList = sasList;

    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

/// @fn         SASDDAPI::getDriveListByController(UINT8 controllerID, vector<SAS_DRIVE_INFO>* returnedList)
/// @brief      list attached SAS drive by specific IOC number
/// @param      IOC ID
/// @param      A pointer to the return device info list
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDDAPI::getDriveListByController(UINT8 controllerID, vector<SAS_DRIVE_INFO>* returnedList)
{
    SasDriveInfo driveInfo;
    string pciBDF;
    CHAR* blockDirToRead = (CHAR *)"/sys/block/";
    DIR* dir = opendir(blockDirToRead);
    struct dirent *dirent;
    struct stat fileAttribute;
    CHAR tempBuffer[256];
    CHAR fileName[256] = "";
    UINT8 driveNum, ramDriveNum;

    if(sasList.empty())
    {
        return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DEVICE_NODE_LIST | ERROR_SAS_DDAPI_NO_SAS_CONTROLLER), NULL, NULL);
    }
    if(dir == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_DEVICE_NODE_LIST | ERROR_OPEN_DIRECTORY), NULL, NULL);
    }
    for(vector<SAS_CONTROLLER_INFO>::iterator iter = sasList.begin(); iter != sasList.end(); iter++)
    {
        driveNum = 0;
        ramDriveNum = 0;

        if (controllerID == (iter - sasList.begin()))
        {
            pciBDF = iter->domindBusDeviceFunction;

            while((dirent = readdir(dir)) != NULL)
            {
                snprintf(fileName, sizeof(fileName), "%s%s", blockDirToRead, dirent->d_name);
                lstat(fileName, &fileAttribute);

                if(S_ISDIR(fileAttribute.st_mode))
                {
                    // This directory dirent is another directory
                }
                else if(S_ISLNK(fileAttribute.st_mode))
                {
                    INT32 linkLength = readlink(fileName, tempBuffer, sizeof(tempBuffer) - 1);
                    if(linkLength != -1)
                    {
                        tempBuffer[linkLength] = '\0';
                    }
                    else
                    {
                        closedir(dir);
                        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_DEVICE_NODE_LIST | ERROR_SAS_DDAPI_READ_LINK_ERROR), NULL, NULL);;
                    }

                    CHAR tempString[256] = "";
                    snprintf(tempString, sizeof(tempString), "%s", pciBDF.c_str());
                    CHAR *chpAddress = strstr(tempBuffer, tempString);

                    if(chpAddress == NULL)
                    {
                        continue;
                    }

                    driveInfo.controllerName = iter->wellKnowName;
                    driveInfo.deviceNodeName = (string)"/dev/" + dirent->d_name;
                    if(strstr(dirent->d_name, "ram"))
                        driveInfo.wellKnownName = iter->wellKnowName + "-RAMDISK" + Int2Str(ramDriveNum++);
                    else
                    {
                        if((m_errorMsg = getDrivePHYInfoByName(&driveInfo))!= NO_ERROR_MSG)
                            return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_DEVICE_NODE_LIST), NULL, m_errorMsg);
                        for(int i = 0 ; i < NUMOFHARDDRIVE ; i++ )
                        {
                            if(driveInfo.sasPhyInfo[0].attachedPhyID == aurora[i])
                            {
                                driveInfo.wellKnownName = iter->wellKnowName + "-DRIVE" + Int2Str(i);
                                break;
                            }
                        }
                    }
                    // start to get SAS address
                    CHAR devicePath[128] = "", tempAddress[20];
                    FILE *fp = NULL;
                    snprintf(devicePath, sizeof(devicePath), "cat /sys/block/%s/device/sas_address", dirent->d_name);

                    fp = popen(devicePath, "r");

                    if(!fp)
                        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_DEVICE_NODE_LIST | ERROR_READ_FILE_FAIL), NULL, NULL);

                    if(fgets(tempAddress, sizeof(tempAddress), fp) != NULL)
                    {
                        if (tempAddress[strlen(tempAddress)-1] == '\n')
                            tempAddress[strlen(tempAddress)-1] = '\0';
                        driveInfo.sasAddress = std::strtoull(tempAddress, NULL, 16);
                    }
                    if(fp)
                        pclose(fp);

                    // start to get SG device node
                    CHAR SGDirToRead[128] = "";
                    snprintf(SGDirToRead, sizeof(SGDirToRead), "/sys/block/%s/device/scsi_generic/", dirent->d_name);
                    DIR *SGDir = opendir(SGDirToRead);

                    if(SGDir == NULL)
                    {
                        closedir(SGDir);
                        closedir(dir);
                        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_DEVICE_NODE_LIST | ERROR_OPEN_DIRECTORY), NULL, NULL);
                    }

                    struct dirent *SGEntry;
                    struct stat SGFileAttribute;

                    while((SGEntry = readdir(SGDir)) != NULL)
                    {
                        CHAR szSGFileName[256];

                        snprintf(szSGFileName, sizeof(szSGFileName), "%s", SGDirToRead);
                        lstat(szSGFileName, &SGFileAttribute);

                        if(S_ISDIR(SGFileAttribute.st_mode))
                        {
                            if(!strcmp(SGEntry->d_name, ".") || !strcmp(SGEntry->d_name, ".."))
                            {
                              continue;
                            }
                            driveInfo.sgDeviceNodeName = (string)"/dev/" + SGEntry->d_name;
                        }
                    }
                    closedir(SGDir);
                    // end to get SG device node
                    driveList.push_back(driveInfo);
                }
                else if(S_ISREG(fileAttribute.st_mode))
                {
                  // This dirent is a regular filename
                }
            }
            closedir(dir);
            break;
        }
    }
    *returnedList = driveList;
    driveList.clear();

    return NO_ERROR_MSG;
}

/// @fn         SASDDAPI::getDriveInfoByName(string wellKnownName, SasDriveInfo* driveInfo)
/// @brief      get SAS drive information by specific drive well known name
/// @param      drive well known name
/// @param      A pointer to the return drive info structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDDAPI::getDriveInfoByName(string wellKnownName, SasDriveInfo* driveInfo)
{
    int len, req_len, avail_len;
    int transport = -1;
    int form_factor = 0;
    int haw_zbc = 0;
    int protect = 0;
    int rpm = 0;
    string tempString;
    m_errorMsg = NO_ERROR_MSG;

    if(driveInfo->wellKnownName == wellKnownName)
    {
        INT32 fileDescriptor = open(driveInfo->deviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
        if(fileDescriptor < 0)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO |
                                              ERROR_STORAGE_DDAPI_OPEN_DEVICE_FAIL), NULL, NULL);

        memset(gBuf, 0, 96);
        req_len = 36;
        if ((m_errorMsg = scsiFactory->scsiStdInquiry(fileDescriptor, gBuf, req_len)) != NO_ERROR_MSG)
        {
            /* Marvell controllers fail on a 36 bytes StdInquiry, but 64 suffices */
            req_len = 64;
            if((m_errorMsg = scsiFactory->scsiStdInquiry(fileDescriptor, gBuf, req_len)) != NO_ERROR_MSG)
                return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);
        }

        avail_len = gBuf[4] + 5;
        len = (avail_len < req_len) ? avail_len : req_len;

        if(len < 36)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), "Short INQUIRY response, skip product id\n", NULL);

        // Upper bits of version bytes were used in older standards
        // Only interested in SPC-4 (0x6) and SPC-5 (assumed to be 0x7)
        //scsi_version = gBuf[2] & 0x7;

        if(0 != strncmp((char *)&gBuf[8], "ATA", 3))
        {
            UCHAR vendor[8+1], product[16+1], revision[4+1];
            scsiFactory->scsiFormatIDtoString((STRING)&gBuf[8], 8, vendor);
            scsiFactory->scsiFormatIDtoString((STRING)&gBuf[16], 16, product);
            scsiFactory->scsiFormatIDtoString((STRING)&gBuf[32], 4, revision);
            driveInfo->vendorName.assign((const char*)vendor);
            driveInfo->productName.assign((const char*)product);
            driveInfo->revision.assign((const char*)revision);
        }

        protect = gBuf[5] & 0x1;    /* from and including SPC-3 */

        unsigned int lb_size = 0;
        int lb_per_pb_exp = 0;
        UINT64 capacity;

        if((m_errorMsg = getDriveSizeByNodeName(fileDescriptor, &lb_size,
                                              &lb_per_pb_exp, &capacity)) != NO_ERROR_MSG)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);

        if(capacity)
        {
            driveInfo->capacity = capacity;
            driveInfo->logicalBlockSize = lb_size;
        }
        if (protect || lb_per_pb_exp)
        {
            unsigned char rc16_12[20] = {0, };

            if(scsiFactory->scsiGetProtectionPhysicalInfo(fileDescriptor, rc16_12) == NO_ERROR_MSG)
            {
                lb_per_pb_exp = rc16_12[1] & 0xf;       /* just in case */
                if (lb_per_pb_exp > 0)
                    driveInfo->physicalBlockSize = lb_size * (1 << lb_per_pb_exp);
            }
        }

        if((m_errorMsg = scsiFactory->getSCSIDriveRPM(fileDescriptor, 0, &rpm,
                                                    &form_factor, &haw_zbc))!= NO_ERROR_MSG)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);

        if(rpm >= 0)
        {
            if (0 == rpm)
                ;       // Not reported
            else if(1 == rpm)
                driveInfo->rotationRate = "Solid State Device";
            else if((rpm <= 0x400) || (0xffff == rpm))
                ;       // Reserved
            else
                driveInfo->rotationRate = Int2Str(rpm) + " rpm";
        }
        if (form_factor > 0)
        {
            const char * cp = NULL;

            switch (form_factor)
            {
                case 1:
                    cp = "5.25";
                    break;
                case 2:
                    cp = "3.5";
                    break;
                case 3:
                    cp = "2.5";
                    break;
                case 4:
                    cp = "1.8";
                    break;
                case 5:
                    cp = "< 1.8";
                    break;
            }
            if (cp)
                driveInfo->formFactor = tempString.assign(cp) + " inches";
        }

        if((m_errorMsg = scsiFactory->scsiInquiryVpd(fileDescriptor, SCSI_VPD_DEVICE_IDENTIFICATION,
                                       gBuf, 252))!= NO_ERROR_MSG)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);

        UCHAR WWID[256];

        len = gBuf[3];
        scsiFactory->scsiDecodeLUDevID((STRING)gBuf + 4, len, WWID, sizeof(WWID), &transport);
        if (strlen((const char *)WWID) > 0)
            driveInfo->logicalUnitID.assign((const char *)WWID);

        if((m_errorMsg = scsiFactory->scsiInquiryVpd(fileDescriptor, SCSI_VPD_UNIT_SERIAL_NUMBER,
                                       gBuf, 252))!= NO_ERROR_MSG)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);

        UCHAR serial[256];
        len = gBuf[3];

        gBuf[4 + len] = '\0';
        scsiFactory->scsiFormatIDtoString((STRING)&gBuf[4], len, serial);
        driveInfo->serialNumber.assign((const char*)serial);

        // See if transport protocol is known
        if(transport < 0)
        {
            if((m_errorMsg = scsiFactory->getSCSITransportProtocol(fileDescriptor, 0, &transport)) != NO_ERROR_MSG)
                return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);
        }
        if ((transport >= 0) && (transport <= 0xf))
            driveInfo->protocalType.assign((const char*)transport_proto_arr[transport]);

        UINT8 currenttemp = 0, triptemp = 0;
        if((m_errorMsg = scsiFactory->getSCSIDriveTemperature(fileDescriptor, &currenttemp, &triptemp)) != NO_ERROR_MSG)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);
        driveInfo->temperature = Int2Str(currenttemp) + " C";

        if((m_errorMsg = getDrivePHYInfoByName(driveInfo)) != NO_ERROR_MSG)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);

        if((m_errorMsg = scsiFactory->getSCSIErrorLog(fileDescriptor, &driveInfo->smartData.errorLog)) != NO_ERROR_MSG)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);

        if((m_errorMsg = scsiFactory->getSmartHealthStatus(fileDescriptor, &driveInfo->smartData.strHealthStatus)) != NO_ERROR_MSG)
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO), NULL, m_errorMsg);
    }
    else
        return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_INFO | ERROR_BAD_PARAMETER), NULL, NULL);

	return m_errorMsg;
}

// See Serial Attached SCSI (SPL-3) (e.g. revision 6g) the Protocol Specific
// log page [0x18]. Returns 0 if ok else FAIL* bitmask.
ErrorMsg*
SASDDAPI::getDrivePHYInfoByName(SasDriveInfo* driveInfo)
{
    int num;
    UINT32 k, param_len, index = 0;
    UCHAR* ucp;

    if(driveInfo->deviceNodeName == "")
        return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_PHY_INFO | ERROR_STORAGE_DDAPI_OPEN_DEVICE_FAIL), NULL, NULL);

    INT32 fileDescriptor = open(driveInfo->deviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
    if(fileDescriptor < 0)
        return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_PHY_INFO | ERROR_STORAGE_DDAPI_OPEN_DEVICE_FAIL), NULL, NULL);
    if((m_errorMsg = scsiFactory->scsiLogSense(fileDescriptor, PROTOCOL_SPECIFIC_LPAGE, 0, gBuf, LOG_RESP_LONG_LEN, 0)))
        return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_PHY_INFO), NULL, m_errorMsg);

    if((gBuf[0] & 0x3f) != PROTOCOL_SPECIFIC_LPAGE)
        return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_PHY_INFO),
                              "Protocol specific Log Sense Failed, page mismatch", NULL);

    // compute page length
    num = (gBuf[2] << 8) + gBuf[3];

    for (k = 0, ucp = gBuf + 4; k < num; )
    {
        param_len = ucp[3] + 4;
        if (6 != (0xf & ucp[4]))
            return returnErrorMsg((ERRORCODE)(ERROR_SAS_DDAPI_GET_DRIVE_PHY_INFO),
                    "Only support protocol specific log page on SAS devices", NULL);   /* only decode SAS log page */
        getDriveSASPortParam(ucp, param_len, &driveInfo->sasPhyInfo[index]);
        k += param_len;
        ucp += param_len;
        index++;
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDDAPI::getInformationByController(UINT8 controllerID, SasInfo* sasInfo)
/// @brief      get SAS IOC information by controller ID
/// @param      controller ID
/// @param      A pointer to the return SAS info structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDDAPI::getInformationByController(UINT8 controllerID, SasInfo* sasInfo)
{
    for(vector<SAS_CONTROLLER_INFO>::iterator iter = sasList.begin(); iter != sasList.end(); iter++)
    {
        if (controllerID == (iter - sasList.begin()))
        {/*
            // start to get firmware info from sas3flash utility
            CHAR command[128] = "", tempString[100];
            FILE *fp = NULL;
            int lineNum = 0;
            snprintf(command, sizeof(command), "sas3flash -c %d -list", furyRealID[controllerID]);

            fp = popen(command, "r");
            if(!fp)
                return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_SAS_INFO | ERROR_READ_FILE_FAIL), NULL, NULL);

            while(fgets(tempString, sizeof(tempString), fp))
            {
                if (tempString[strlen(tempString)-1] == '\n')
                    tempString[strlen(tempString)-1] = '\0';
                if(lineNum == 10)
                    iter->defaultNVDATAVersion = iter->defaultNVDATAVersion.assign(tempString).substr(34);
                if(lineNum == 11)
                    iter->persistentNVDATAVersion = iter->persistentNVDATAVersion.assign(tempString).substr(34);
                if(lineNum == 12)
                    iter->firmwarePID = iter->firmwarePID.assign(tempString).substr(34);
                if(lineNum == 13)
                    iter->firmwareVersion = iter->firmwareVersion.assign(tempString).substr(34);
                if(lineNum ==14)
                    iter->nvdataVendor = iter->nvdataVendor.assign(tempString).substr(34);
                if(lineNum ==15)
                    iter->nvdataPID = iter->nvdataPID.assign(tempString).substr(34);
                if(lineNum ==16)
                    iter->biosVersion = iter->biosVersion.assign(tempString).substr(34);
                if(lineNum == 17)
                    iter->uefiBSDVersion = iter->uefiBSDVersion.assign(tempString).substr(34);
                if(lineNum ==18)
                    iter->fcodeVersion = iter->fcodeVersion.assign(tempString).substr(34);

                lineNum++;
            }
            if(fp)
                pclose(fp);
*/
            *sasInfo = *iter;

            break;
        }
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDDAPI::programImage
/// @brief      Program an image to the requested SAS device
/// @param      IOC ID
/// @param      image buffer
/// @param      image buffer size
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDDAPI::programImage(UINT8 controllerID, UINT8* imageBuffer, UINT32 imageBufferSize)
{
    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

/// @fn         SASDDAPI::setControllerID
/// @brief      set controller ID for SAS driver
/// @param      IOC ID
/// @exception  Nothing
/// @return     None
void
SASDDAPI::setControllerID(UINT8 controllerID)
{
    sasIOCTL->setControllerID(furyRealID[controllerID]);
}

/// @fn         SASDDAPI::setControllerID
/// @brief      set controller ID for SAS driver
/// @param      IOC ID
/// @exception  Nothing
/// @return     None
void
SASDDAPI::setPHYNumber(UINT8 phyNumber)
{
    sasIOCTL->setPHYNumber(phyNumber);
}

/* Return number of bytes of storage in 'device' or 0 if error. If
 * successful and lb_sizep is not NULL then the logical block size
 * in bytes is written to the location pointed to by lb_sizep. */
ErrorMsg*
SASDDAPI::getDriveSizeByNodeName(INT32 fileDescription, UINT32* logicalBlockSize, int* lb_per_pb_expp, UINT64* driveSize)
{
    unsigned int last_lba = 0, lb_size = 0;
    int k;
    uint64_t ret_val = 0;
    UINT8 rc16resp[32];

    m_errorMsg = scsiFactory->scsiReadCapacity10(fileDescription, &last_lba, &lb_size);
    if(m_errorMsg != (ErrorMsg *) NO_ERROR)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_DRIVE_SIZE), NULL, m_errorMsg);
    }

    if (0xffffffff == last_lba)
    {
        m_errorMsg = scsiFactory->scsiReadCapacity16(fileDescription, rc16resp, sizeof(rc16resp));
        if(m_errorMsg != (ErrorMsg *) NO_ERROR)
        {
            return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_DRIVE_SIZE), NULL, m_errorMsg);
        }

        for (k = 0; k < 8; ++k)
        {
            if (k > 0)
                ret_val <<= 8;
            ret_val |= rc16resp[k + 0];
        }
        if (lb_per_pb_expp)
            *lb_per_pb_expp = (rc16resp[13] & 0xf);
    }
    else
    {
        ret_val = last_lba;
        if (lb_per_pb_expp)
            *lb_per_pb_expp = 0;
    }
    if (logicalBlockSize)
        *logicalBlockSize = lb_size;
    ++ret_val;  /* last_lba is origin 0 so need to bump to get number of */

    if(!(*driveSize = ret_val * lb_size))
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DDAPI_GET_DRIVE_SIZE), NULL, m_errorMsg);

    return NO_ERROR_MSG;
}

void
SASDDAPI::getDriveSASPortParam(UCHAR* ucp, UINT32 param_len, SasDrivePHYInfo* drivePhyInfo)
{
    int j, n, t, sz, spld_len;
    unsigned char * vcp;
    UINT64 ull;
    unsigned int ui;
    char s[64];

    sz = sizeof(s);
    t = (ucp[0] << 8) | ucp[1];
    drivePhyInfo->portID = t;

    for (j = 0, vcp = ucp + 8; j < (param_len - 8); vcp += spld_len, j += spld_len)
    {
        drivePhyInfo->phyID = vcp[1];
        spld_len = vcp[3];
        if (spld_len < 44)
            spld_len = 48;      /* in SAS-1 and SAS-1.1 vcp[3]==0 */
        else
            spld_len += 4;
        t = ((0x70 & vcp[4]) >> 4);
        switch (t)
        {
            case 0: snprintf(s, sz, "no device attached"); break;
            case 1: snprintf(s, sz, "SAS or SATA device"); break;
            case 2: snprintf(s, sz, "expander device"); break;
            case 3: snprintf(s, sz, "expander device (fanout)"); break;
            default: snprintf(s, sz, "reserved [%d]", t); break;
        }
        drivePhyInfo->attachedType.assign(s);

        t = (0xf & vcp[5]);
        switch (t)
        {
            case 0: snprintf(s, sz, "phy enabled; unknown");
                         break;
            case 1: snprintf(s, sz, "phy disabled"); break;
            case 2: snprintf(s, sz, "speed negotiation failed");
                         break;
            case 3: snprintf(s, sz, "SATA spinup hold state");
                         break;
            case 4: snprintf(s, sz, "port selector");
                         break;
            case 5: snprintf(s, sz, "reset in progress");
                         break;
            case 6: snprintf(s, sz, "unsupported phy attached");
                         break;
            case 8: snprintf(s, sz, "1.5 Gbps"); break;
            case 9: snprintf(s, sz, "3 Gbps"); break;
            case 0xa: snprintf(s, sz, "6 Gbps"); break;
            case 0xb: snprintf(s, sz, "12 Gbps"); break;
            default: snprintf(s, sz, "reserved [%d]", t); break;
        }
        drivePhyInfo->linkRate.assign(s);

        for (n = 0, ull = vcp[8]; n < 8; ++n)
        {
            ull <<= 8; ull |= vcp[8 + n];
        }
        drivePhyInfo->sasAddress = ull;

        for (n = 0, ull = vcp[16]; n < 8; ++n)
        {
            ull <<= 8; ull |= vcp[16 + n];
        }
        drivePhyInfo->attachedSASAddress =ull;

        drivePhyInfo->attachedPhyID = vcp[24];
        ui = (vcp[32] << 24) | (vcp[33] << 16) | (vcp[34] << 8) | vcp[35];
        drivePhyInfo->InvalidDword = ui;
        ui = (vcp[36] << 24) | (vcp[37] << 16) | (vcp[38] << 8) | vcp[39];
        drivePhyInfo->RunningDisparity = ui;
        ui = (vcp[40] << 24) | (vcp[41] << 16) | (vcp[42] << 8) | vcp[43];
        drivePhyInfo->LossOfDwordSync = ui;
        ui = (vcp[44] << 24) | (vcp[45] << 16) | (vcp[46] << 8) | vcp[47];
        drivePhyInfo->PhyResetProblem = ui;
    }
}

bool
SASDDAPI::sortSAScontroller(const SAS_CONTROLLER_INFO &left, const SAS_CONTROLLER_INFO &right)
{
    return left.wellKnowName < right.wellKnowName;
}
