/// @file       SASDDAPI.hpp
/// @brief      SASDDAPI class
/// @details    The header file of SAS DDAPI
/// @version    1.0.0
/// @date       2014.01.06
/// @author     Jerry Lee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.
#ifndef SASDDAPI_H
#define SASDDAPI_H

#include <fcntl.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <unistd.h>
#include <strings.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <limits.h>
#include <dirent.h>
#include <vector>
#include <linux/stddef.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <dirent.h>
#include <scsi/sg.h>

#include "BaseClassDDAPI.hpp"
#include "SASIOCTL.hpp"
#include "PCIDDAPI.hpp"
#include "SCSIFactory.hpp"
using std::vector;

static unsigned char aurora[] = {
    16, 17, 18, 19, 20, 21, 22, 23,
    24, 25, 26, 27, 28, 29, 30, 31,
    32, 33, 34, 35, 36, 37, 38, 39
};

static unsigned char sasControllerID[] = {
    1,2,0
};

static unsigned char sasControllerIDwithNoHBA[] = {
    1,0
};

static unsigned char furyRealID[] = {
    1,0
};

enum SAS_REQUEST_STRUCT {
    EXECUTE_CONTROL,
    LOOPBACK_PARAMETER,
    NULL_PARAMETER,
};

// SAS information definitions
typedef struct SAS_CONTROLLER_INFO
{
    string wellKnowName;
    string driverVersion;
    string firmwareVersion;
    string biosVersion;
    string domindBusDeviceFunction;
    string hardwareRevision;
    string defaultNVDATAVersion;
    string persistentNVDATAVersion;
    string firmwarePID;
    string nvdataVendor;
    string nvdataPID;
    string uefiBSDVersion;
    string fcodeVersion;
    float  pciSpeed;
    UINT32 pciWidth;

} SasInfo;

// SAS drive PHY information
typedef struct SAS_DRIVE_PHY_INFO
{
        UINT8 portID;
        UINT8 phyID;
        UINT8 attachedPhyID;
        string attachedType;
        string linkRate;
        UINT64 sasAddress;
        UINT64 attachedSASAddress;
        UINT32 InvalidDword;
        UINT32 RunningDisparity;
        UINT32 LossOfDwordSync;
        UINT32 PhyResetProblem;
} SasDrivePHYInfo;

// SAS drive node information
typedef struct SAS_DRIVE_INFO
{
    string wellKnownName;
    string controllerName;
    string deviceNodeName;
    string sgDeviceNodeName;
    string vendorName;
    string productName;
    string revision;
    string formFactor;
    string rotationRate;
    UINT64 capacity;
    UINT64 sasAddress;
    UINT32 logicalBlockSize;
    UINT32 physicalBlockSize;
    string serialNumber;
    string logicalUnitID;
    string temperature;
    string protocalType;
    SasDrivePHYInfo sasPhyInfo[2];
    StorageSASSMARTDataStruct smartData;
} SasDriveInfo;


class SASDDAPI: public BaseClassDDAPI
{
    public:
        SASDDAPI(STRING className);
        virtual ~SASDDAPI(void);

        ErrorMsg* doIOCTLCommand(SASIOCTL::IOCTL_OP_CODE ioctlCommand, void* returnedValue);
        ErrorMsg* getControllerList(vector<SAS_CONTROLLER_INFO>* returnedList);
        ErrorMsg* getDriveListByController(UINT8 controllerID, vector<SAS_DRIVE_INFO>* returnedList);
        ErrorMsg* getDriveInfoByName(string wellKnownName, SasDriveInfo* driveInfo);
        ErrorMsg* getDrivePHYInfoByName(SasDriveInfo* driveInfo);
        ErrorMsg* getInformationByController(UINT8 controllerID, SasInfo* sasInfo);
        ErrorMsg* getPHYCounts(UINT8 controllerID, UINT8* returnCounts);
        ErrorMsg* programImage(UINT8 controllerID, UINT8* imageBuffer, UINT32 imageBufferSize);
        void setControllerID(UINT8 controllerID);
        void setPHYNumber(UINT8 phyNumber);
        void setLoopbackParam(UINT32 sourceLogicalID, UINT32 destLogicalID, UINT32 loopbackPath, UINT32 numberOfBlocks,
                              UINT32 blockSize, UINT32 testPattern, UINT32 action);

    protected:
	
    private:
        IOCTL* sasIOCTL;
        PCIDDAPI* pciDDAPI;
        SCSIFactory* scsiFactory;
        sas_loopback_param loopbackParam;
        vector<SAS_CONTROLLER_INFO> sasList;
        vector<ERROR_COUNT_INFO> errorList;
        vector<SAS_DRIVE_INFO> driveList;
        ErrorMsg* m_errorMsg = (ErrorMsg *)NO_ERROR;
        UINT8 gBuf[65535];

        ErrorMsg* addIOCController(void);
        ErrorMsg* addDriveInfo(void);

        ErrorMsg* getDriveSizeByNodeName(INT32 fileDescription, UINT32* logicalBlockSize, int* lb_per_pb_expp, UINT64* driveSize);
        // get drive SAS port parameters
        void getDriveSASPortParam(UCHAR* ucp, UINT32 param_len, SasDrivePHYInfo* drivePhyInfo);
        static bool sortSAScontroller(const SAS_CONTROLLER_INFO &left, const SAS_CONTROLLER_INFO &right);
};


#endif // SASDDAPI_H

