/// @file       SCSISenseAnalyzer.cpp
/// @brief      SCSISenseAnalyzer class
/// @details    Provide low layer diagnostic and utility APIs for Storage module
/// @version    1.0.0
/// @date       2015.12.01
/// @author     Ben Yang
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.
#include "SCSISenseAnalyzer.hpp"

/// @fn         SCSISenseAnalyzer::SCSISenseAnalyzer(STRING in_strClassName)
/// @brief      SCSISenseAnalyzer constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
SCSISenseAnalyzer::SCSISenseAnalyzer(STRING in_szClassName) : BaseClassDDAPI(in_szClassName)
{
     m_strOutputResultString = "";
     m_blsSenseKeyTableInit = 0;
}

/// @fn         SCSISense::~SCSISense(void)
/// @brief      SCSISense destructor
/// @exception  Nothing
/// @return     Nothing
SCSISenseAnalyzer::~SCSISenseAnalyzer(void)
{
     m_strOutputResultString = "";
     m_blsSenseKeyTableInit = 0;
}

/// @fn         ErrorMsg *SCSISenseAnalyzer::addSenseDataByID
/// @brief      list Storage controller's ports
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SCSISenseAnalyzer::setSenseDataByIndex(UINT32 in_u32DefinedIndex, UCHAR* in_uchaSenseBuffer)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     UINT32 u32SenseKey = 0;
     UINT32 u32AdditionalSenseCode = 0;
     UINT32 u32AdditionalSenseCodeQualifier = 0;
     if((errorCode = parseSenseData(in_uchaSenseBuffer, &u32SenseKey, &u32AdditionalSenseCode, &u32AdditionalSenseCodeQualifier)) != NO_ERROR)
     {
	  return returnErrorMsg(errorCode, NULL, NULL);
     }
     if(u32SenseKey == 0)
     {
	  return returnErrorMsg(errorCode, NULL, NULL);
     }

     BOOL blsFoundDefinedIndex = 0;
     list <SCSISenseDriveStruct> *pDriveSenseBufferList = &m_DriveList;
     for(list <SCSISenseDriveStruct>::iterator driveListIterator = pDriveSenseBufferList->begin(); driveListIterator != pDriveSenseBufferList->end(); driveListIterator++)
     {
	  if(driveListIterator->u32DefinedIndex == in_u32DefinedIndex)
	  {
	       blsFoundDefinedIndex = 1;
	  }
     }

     if(blsFoundDefinedIndex) // if found definedIndex then set data
     {
	  for(list <SCSISenseDriveStruct>::iterator driveListIterator = pDriveSenseBufferList->begin(); driveListIterator != pDriveSenseBufferList->end(); driveListIterator++)
	  {
	       if(driveListIterator->u32DefinedIndex == in_u32DefinedIndex)
	       {
		    for(list <SCSISenseDataStruct>::iterator senseListIterator = driveListIterator->SCSISenseDataList.begin(); senseListIterator != driveListIterator->SCSISenseDataList.end(); senseListIterator++)
		    {
			 // update old sense data count
			 if(senseListIterator->u32SenseKey==u32SenseKey && senseListIterator->u32AdditionalSenseCode==u32AdditionalSenseCode && senseListIterator->u32AdditionalSenseCodeQualifier==u32AdditionalSenseCodeQualifier)
			 {
			      senseListIterator->u32ErrorCount++;
			      return returnErrorMsg(errorCode, NULL, NULL);
			 }
		    }

		    // add new sense data
		    initSenseKeyTable();
		    SCSISenseDataStruct tempSCSISenseData;
		    tempSCSISenseData.strCategory = m_SCSISenseKeyTable[u32SenseKey][u32AdditionalSenseCode][u32AdditionalSenseCodeQualifier].strCategory;
		    tempSCSISenseData.strErrorCondition = m_SCSISenseKeyTable[u32SenseKey][u32AdditionalSenseCode][u32AdditionalSenseCodeQualifier].strErrorCondition;

		    tempSCSISenseData.u32SenseKey = u32SenseKey;
		    tempSCSISenseData.u32AdditionalSenseCode = u32AdditionalSenseCode;
		    tempSCSISenseData.u32AdditionalSenseCodeQualifier = u32AdditionalSenseCodeQualifier;
		    tempSCSISenseData.u32ErrorCount = 1;
		    driveListIterator->SCSISenseDataList.push_back(tempSCSISenseData);
		    return returnErrorMsg(errorCode, NULL, NULL);
	       }
	  }
     }
     else // definedIndex is not found, add new data to m_DriveList and new sense data
     {
	  // add drive sense buffer
	  SCSISenseDriveStruct tempSCSISenseDrive;
	  tempSCSISenseDrive.u32ControllerType = 0;
	  tempSCSISenseDrive.u32DefinedIndex = in_u32DefinedIndex;
	  tempSCSISenseDrive.strDriveDefinedName = "testDefinedName";
	  pDriveSenseBufferList->push_back(tempSCSISenseDrive);

	  list <SCSISenseDriveStruct>::iterator driveListIterator = pDriveSenseBufferList->end();
	  driveListIterator--;

	  // add sense data
	  initSenseKeyTable();
	  SCSISenseDataStruct tempSCSISenseData;
	  tempSCSISenseData.strCategory = m_SCSISenseKeyTable[u32SenseKey][u32AdditionalSenseCode][u32AdditionalSenseCodeQualifier].strCategory;
	  tempSCSISenseData.strErrorCondition = m_SCSISenseKeyTable[u32SenseKey][u32AdditionalSenseCode][u32AdditionalSenseCodeQualifier].strErrorCondition;
	  tempSCSISenseData.u32SenseKey = u32SenseKey;
	  tempSCSISenseData.u32AdditionalSenseCode = u32AdditionalSenseCode;
	  tempSCSISenseData.u32AdditionalSenseCodeQualifier = u32AdditionalSenseCodeQualifier;
	  tempSCSISenseData.u32ErrorCount = 1;
	  driveListIterator->SCSISenseDataList.push_back(tempSCSISenseData);

	  return returnErrorMsg(errorCode, NULL, NULL);
     }
     return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DDAPI_GET_DEFINED_INDEX_BY_NAME | errorCode), NULL, NULL);
}

ErrorMsg *SCSISenseAnalyzer::listSenseDataByIndex(UINT32 in_u32DefinedIndex, BOOL in_blsMultiModeOutputFormat)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
     
     stringstream ssTemp;
     ssTemp.clear();
     ssTemp.str("");

     if(in_blsMultiModeOutputFormat == true)
     {
	  list <SCSISenseDriveStruct> *pDriveSenseBufferList = &m_DriveList;
	  for(list <SCSISenseDriveStruct>::iterator driveListIterator = pDriveSenseBufferList->begin(); driveListIterator != pDriveSenseBufferList->end(); driveListIterator++)
	  {
	       if(driveListIterator->u32DefinedIndex == in_u32DefinedIndex)
	       {
		    UINT32 u32NoAnyData = 1;
		    errorCode = NO_ERROR;

		    for(UINT32 u32I=0; u32I<15; u32I++)
		    {
			 for(UINT32 u32J=0; u32J<255; u32J++)
			 {
			      for(UINT32 u32K=0; u32K<255; u32K++)
			      {
				   for(list <SCSISenseDataStruct>::iterator senseListIterator = driveListIterator->SCSISenseDataList.begin(); senseListIterator != driveListIterator->SCSISenseDataList.end(); senseListIterator++)
				   {
					if(senseListIterator->u32SenseKey==m_SCSISenseKeyTable[u32I][u32J][u32K].u32SenseKey &&
					   senseListIterator->u32AdditionalSenseCode==m_SCSISenseKeyTable[u32I][u32J][u32K].u32AdditionalSenseCode &&
					   senseListIterator->u32AdditionalSenseCodeQualifier==m_SCSISenseKeyTable[u32I][u32J][u32K].u32AdditionalSenseCodeQualifier)
					{
					     u32NoAnyData = 0;
					     ssTemp << hex << senseListIterator->u32SenseKey;
					     ssTemp << hex << senseListIterator->u32AdditionalSenseCode;
					     ssTemp << hex << senseListIterator->u32AdditionalSenseCodeQualifier << ":";
					     ssTemp << hex << senseListIterator->u32ErrorCount << ", ";
					}
				   }
			      }
			 }
		    }

		    if(u32NoAnyData == 1)
		    {
			 ssTemp << "NoAnyErrorSenseData\t";
		    }
		    else
		    {
			 errorCode = ERROR_STORAGE_DDAPI_SCSICDBSEND_SENSE_ERROR;
		    }
		    m_strOutputResultString = ssTemp.str();
		    return returnErrorMsg(errorCode, NULL, NULL);
	       }
	  }
	  ssTemp << "NoAnyErrorSenseData\t";
	  m_strOutputResultString = ssTemp.str();
     }
     else
     {
	  ssTemp << endl;
	  ssTemp << "Category" << "\tKey" << "\tASC" << "\tASCQ" << "\tError_Condition" << "\t\tCount" << endl;
	  ssTemp << "--------------------------------------------------------------------------------------------" << endl;

	  list <SCSISenseDriveStruct> *pDriveSenseBufferList = &m_DriveList;
	  for(list <SCSISenseDriveStruct>::iterator driveListIterator = pDriveSenseBufferList->begin(); driveListIterator != pDriveSenseBufferList->end(); driveListIterator++)
	  {
	       if(driveListIterator->u32DefinedIndex == in_u32DefinedIndex)
	       {
		    UINT32 u32NoAnyData = 1;
		    errorCode = NO_ERROR;

		    for(UINT32 u32I=0; u32I<15; u32I++)
		    {
			 for(UINT32 u32J=0; u32J<255; u32J++)
			 {
			      for(UINT32 u32K=0; u32K<255; u32K++)
			      {
				   for(list <SCSISenseDataStruct>::iterator senseListIterator = driveListIterator->SCSISenseDataList.begin(); senseListIterator != driveListIterator->SCSISenseDataList.end(); senseListIterator++)
				   {
					if(senseListIterator->u32SenseKey==m_SCSISenseKeyTable[u32I][u32J][u32K].u32SenseKey &&
					   senseListIterator->u32AdditionalSenseCode==m_SCSISenseKeyTable[u32I][u32J][u32K].u32AdditionalSenseCode &&
					   senseListIterator->u32AdditionalSenseCodeQualifier==m_SCSISenseKeyTable[u32I][u32J][u32K].u32AdditionalSenseCodeQualifier)
					{
					     u32NoAnyData = 0;
					     ssTemp << senseListIterator->strCategory << "\t";
					     ssTemp << hex << senseListIterator->u32SenseKey << "\t";
					     ssTemp << hex << senseListIterator->u32AdditionalSenseCode << "\t";
					     ssTemp << hex << senseListIterator->u32AdditionalSenseCodeQualifier << "\t";
					     ssTemp << senseListIterator->strErrorCondition << "\t\t";
					     ssTemp << hex << senseListIterator->u32ErrorCount << endl;
					}
				   }
			      }
			 }
		    }

		    if(u32NoAnyData == 1)
		    {
			 ssTemp << "NoAnyErrorSenseData" << endl;
		    }
		    else
		    {
			 errorCode = ERROR_STORAGE_DDAPI_SCSICDBSEND_SENSE_ERROR;
		    }
		    ssTemp << endl;
		    m_strOutputResultString = ssTemp.str();
		    return returnErrorMsg(errorCode, NULL, NULL);
	       }
	  }
	  ssTemp << "NoAnyErrorSenseData" << endl;
	  ssTemp << endl;
	  m_strOutputResultString = ssTemp.str();
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SCSISenseAnalyzer::clearSenseDataByIndex(UINT32 in_u32DefinedIndex)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     list <SCSISenseDriveStruct> *pDriveSenseBufferList = &m_DriveList;
     for(list <SCSISenseDriveStruct>::iterator driveListIterator = pDriveSenseBufferList->begin(); driveListIterator != pDriveSenseBufferList->end(); driveListIterator++)
     {
	  if(driveListIterator->u32DefinedIndex == in_u32DefinedIndex)
	  {
	       errorCode = NO_ERROR;
	       driveListIterator->SCSISenseDataList.clear();
	       return returnErrorMsg(errorCode, NULL, NULL);
	  }
     }
     return returnErrorMsg((ERRORCODE)(ERROR_STORAGE_DDAPI_GET_DEFINED_INDEX_BY_NAME | errorCode), NULL, NULL);
}

ErrorMsg *SCSISenseAnalyzer::setSenseDataByName(string in_strDefinedName, UCHAR in_uchaSenseBuffer)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
     return NO_ERROR_MSG;
}

ErrorMsg *SCSISenseAnalyzer::listSenseDataByName(string in_strDefinedName)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
     return NO_ERROR_MSG;
}

ErrorMsg *SCSISenseAnalyzer::clearSenseDataByName(string in_strDefinedName)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;
     return NO_ERROR_MSG;
}

ERRORCODE SCSISenseAnalyzer::translateSenseDataToErrorCode(UCHAR *in_uchpSenseBuffer, UINT32 in_u32BufferSize)
{
    UINT32 u32ResponseCode;
    UINT32 u32SenseKeyValue = 0;
    UINT32 u32AdditionalSenseCode = 0;
    UINT32 u32AdditionalSenseCodeQualifier = 0;

    u32ResponseCode = (in_uchpSenseBuffer[0] & 0x7f);

    if(u32ResponseCode >= 0x72)
    {
        u32SenseKeyValue = (in_uchpSenseBuffer[1] & 0xf);
        u32AdditionalSenseCode = in_uchpSenseBuffer[2];
        u32AdditionalSenseCodeQualifier = in_uchpSenseBuffer[3];
    }
    else if(u32ResponseCode >= 0x70)
    {
        u32SenseKeyValue = (in_uchpSenseBuffer[2] & 0xf);
        if(in_u32BufferSize > 13)
        {
            u32AdditionalSenseCode = in_uchpSenseBuffer[12];
            u32AdditionalSenseCodeQualifier = in_uchpSenseBuffer[13];
        }
    }

    switch (u32SenseKeyValue)
    {
        case SCSI_SK_NO_SENSE:
        case SCSI_SK_RECOVERED_ERR:
            return NO_ERROR;
        case SCSI_SK_NOT_READY:
            if (SCSI_ASC_NO_MEDIUM == u32AdditionalSenseCode)
                return ERROR_SCSI_FACTORY_ERR_NO_MEDIUM;
            else if (SCSI_ASC_NOT_READY == u32AdditionalSenseCode) {
                if (0x1 == u32AdditionalSenseCodeQualifier)
                    return ERROR_SCSI_FACTORY_ERR_BECOMING_READY;
                else
                    return ERROR_SCSI_FACTORY_ERR_NOT_READY;
            } else
                return ERROR_SCSI_FACTORY_ERR_NOT_READY;
        case SCSI_SK_MEDIUM_ERROR:
        case SCSI_SK_HARDWARE_ERROR:
            return ERROR_SCSI_FACTORY_ERR_MEDIUM_HARDWARE;
        case SCSI_SK_ILLEGAL_REQUEST:
            if (SCSI_ASC_UNKNOWN_OPCODE == u32AdditionalSenseCode)
                return ERROR_SCSI_FACTORY_ERR_BAD_OPCODE;
            else if (SCSI_ASC_INVALID_FIELD == u32AdditionalSenseCode)
                return ERROR_SCSI_FACTORY_ERR_BAD_FIELD;
            else if (SCSI_ASC_UNKNOWN_PARAM == u32AdditionalSenseCode)
                return ERROR_SCSI_FACTORY_ERR_BAD_PARAM;
            else
                return ERROR_SCSI_FACTORY_ERR_BAD_PARAM;    /* all other illegal request */
        case SCSI_SK_UNIT_ATTENTION:
            return ERROR_SCSI_FACTORY_ERR_TRY_AGAIN;
        case SCSI_SK_ABORTED_COMMAND:
            return ERROR_SCSI_FACTORY_ERR_ABORTED_COMMAND;
        default:
            return ERROR_SCSI_FACTORY_ERR_UNKNOWN;
    }
}

ERRORCODE SCSISenseAnalyzer::parseSenseData(UCHAR* in_uchaSenseBuffer, UINT32* out_u32pSenseKey, UINT32* out_u32pAdditionalSenseCode, UINT32* out_u32pAdditionalSenseCodeQualifier)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     for(UINT32 u32CNT=0; u32CNT<MAX_SENSE_BUFFER_LENGTH; u32CNT++)
     {
          if(in_uchaSenseBuffer[u32CNT] != 0)
          {
	       // get sense key
	       ssTemp << (in_uchaSenseBuffer[2] & 0xf);
	       ssTemp >> *out_u32pSenseKey;
	       
	       // get attitional sense code
	       ssTemp.str("");
	       ssTemp.clear();
	       ssTemp << (in_uchaSenseBuffer[12] & 0xff);
	       ssTemp >> *out_u32pAdditionalSenseCode;

	       // get attitional sense code qualifier
	       ssTemp.str("");
	       ssTemp.clear();
	       ssTemp << (in_uchaSenseBuffer[13] & 0xff);
	       ssTemp >> *out_u32pAdditionalSenseCodeQualifier;
	       break;
          }
     }
/*
     if(*out_u32pSenseKey != 0)
     {
	  printf("Status: CC\n");
	  cout << "SK:     " << straSenseKeyContent[*out_u32pSenseKey] << endl;
	  printf("SKV:    %02X\n", *out_u32pSenseKey);
	  printf("ASC:    %02X\n", *out_u32pAdditionalSenseCode);
	  printf("ASCQ:   %02X\n", *out_u32pAdditionalSenseCodeQualifier);
	  errorCode = ERROR_STORAGE_DDAPI_SCSICDBSEND_SENSE_ERROR;
	  return errorCode;
     }
*/
     return NO_ERROR;
}

void SCSISenseAnalyzer::initSenseKeyTable(void)
{
     if(m_blsSenseKeyTableInit == 1)
     {
	  return;
     }

     m_blsSenseKeyTableInit = 1;

     for(UINT32 u32I=0; u32I<15; u32I++)
     {
	  for(UINT32 u32J=0; u32J<255; u32J++)
	  {
	       for(UINT32 u32K=0; u32K<255; u32K++)
	       {
		    m_SCSISenseKeyTable[u32I][u32J][u32K].strErrorCondition = "N/A";
		    m_SCSISenseKeyTable[u32I][u32J][u32K].u32SenseKey = u32I;
		    m_SCSISenseKeyTable[u32I][u32J][u32K].u32AdditionalSenseCode = u32J;
		    m_SCSISenseKeyTable[u32I][u32J][u32K].u32AdditionalSenseCodeQualifier = u32K;
		    m_SCSISenseKeyTable[u32I][u32J][u32K].u32ErrorCount = 0;
		    switch(u32I)
		    {
		    case 0:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "NO Sense";
			 break;
		    case 1:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Soft Error";
			 break;
		    case 2:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Not Ready";
			 break;
		    case 3:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Medium Error";
			 break;
		    case 4:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Hardware Error";
			 break;
		    case 5:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Illegal Request";
			 break;
		    case 6:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Unit Attention";
			 break;
		    case 7:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Write Protect";
			 break;
		    case 0xb:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Aborted Command";
			 break;
		    default:
			 m_SCSISenseKeyTable[u32I][u32J][u32K].strCategory = "Other";
			 break;
		    }
	       }
	  }
	  m_SCSISenseKeyTable[u32I][0x5][0x0].strErrorCondition = "Illegal request";
	  m_SCSISenseKeyTable[u32I][0x6][0x0].strErrorCondition = "Unit attention";
	  m_SCSISenseKeyTable[u32I][0x7][0x0].strErrorCondition = "Data protect";
	  m_SCSISenseKeyTable[u32I][0x8][0x0].strErrorCondition = "LUN communication failure";
	  m_SCSISenseKeyTable[u32I][0x8][0x1].strErrorCondition = "LUN communication timeout";
	  m_SCSISenseKeyTable[u32I][0x8][0x2].strErrorCondition = "LUN communication parity error";
	  m_SCSISenseKeyTable[u32I][0x8][0x3].strErrorCondition = "LUN communication CRC error";
	  m_SCSISenseKeyTable[u32I][0x9][0x0].strErrorCondition = "vendor specific sense key";
	  m_SCSISenseKeyTable[u32I][0x9][0x1].strErrorCondition = "servo fault";
	  m_SCSISenseKeyTable[u32I][0x9][0x4].strErrorCondition = "head select fault";
	  m_SCSISenseKeyTable[u32I][0x0A][0x0].strErrorCondition = "error log overflow";
	  m_SCSISenseKeyTable[u32I][0x0B][0x0].strErrorCondition = "aborted command";
	  m_SCSISenseKeyTable[u32I][0x0C][0x0].strErrorCondition = "write error";
	  m_SCSISenseKeyTable[u32I][0x0C][0x2].strErrorCondition = "write error - auto-realloc failed";
	  m_SCSISenseKeyTable[u32I][0x0E][0x0].strErrorCondition = "data miscompare";
	  m_SCSISenseKeyTable[u32I][0x12][0x0].strErrorCondition = "address mark not found for ID field";
	  m_SCSISenseKeyTable[u32I][0x14][0x0].strErrorCondition = "logical block not found";
	  m_SCSISenseKeyTable[u32I][0x15][0x0].strErrorCondition = "random positioning error";
	  m_SCSISenseKeyTable[u32I][0x15][0x1].strErrorCondition = "mechanical positioning error";
	  m_SCSISenseKeyTable[u32I][0x15][0x2].strErrorCondition = "positioning error detected by read of medium";
	  m_SCSISenseKeyTable[u32I][0x27][0x0].strErrorCondition = "write protected";
	  m_SCSISenseKeyTable[u32I][0x29][0x0].strErrorCondition = "POR or bus reset occurred";
	  m_SCSISenseKeyTable[u32I][0x31][0x1].strErrorCondition = "format failed";
	  m_SCSISenseKeyTable[u32I][0x32][0x1].strErrorCondition = "defect list update error";
	  m_SCSISenseKeyTable[u32I][0x32][0x2].strErrorCondition = "no spares available";
	  m_SCSISenseKeyTable[u32I][0x35][0x1].strErrorCondition = "unspecified enclosure services failure";
	  m_SCSISenseKeyTable[u32I][0x37][0x0].strErrorCondition = "parameter rounded";
	  m_SCSISenseKeyTable[u32I][0x3D][0x0].strErrorCondition = "invalid bits in identify message";
	  m_SCSISenseKeyTable[u32I][0x3E][0x0].strErrorCondition = "LUN not self-configured yet";
	  m_SCSISenseKeyTable[u32I][0x40][0x1].strErrorCondition = "DRAM parity error";
	  m_SCSISenseKeyTable[u32I][0x40][0x2].strErrorCondition = "DRAM parity error";
	  m_SCSISenseKeyTable[u32I][0x42][0x0].strErrorCondition = "power-on or self-test failure";
	  m_SCSISenseKeyTable[u32I][0x4C][0x0].strErrorCondition = "LUN failed self-configuration";
	  m_SCSISenseKeyTable[u32I][0x5C][0x0].strErrorCondition = "RPL status change";
	  m_SCSISenseKeyTable[u32I][0x5C][0x1].strErrorCondition = "spindles synchronised";
	  m_SCSISenseKeyTable[u32I][0x5C][0x2].strErrorCondition = "spindles not synchronised";
	  m_SCSISenseKeyTable[u32I][0x65][0x0].strErrorCondition = "voltage fault";
     }
     for(UINT32 u32I=0; u32I<15; u32I++)
     {
	  for(UINT32 u32J=80; u32J<255; u32J++)
	  {
	       for(UINT32 u32K=0; u32K<255; u32K++)
	       {
		    m_SCSISenseKeyTable[u32I][u32J][u32K].strErrorCondition = "Vendor specific";
	       }
	  }
	  for(UINT32 u32K=80; u32K<255; u32K++)
	  {
	       for(UINT32 u32J=0; u32J<255; u32J++)
	       {
		    m_SCSISenseKeyTable[u32I][u32J][u32K].strErrorCondition = "Vendor specific";
	       }
	  }
     }

     m_SCSISenseKeyTable[0x0][0x0][0x0].strErrorCondition = "No error";
     m_SCSISenseKeyTable[0x0][0x5D][0x0].strErrorCondition = "No sense - PFA threshold reached";
     m_SCSISenseKeyTable[0x1][0x1][0x0].strErrorCondition = "Recovered Write error - no index";
     m_SCSISenseKeyTable[0x1][0x2][0x0].strErrorCondition = "Recovered no seek completion";
     m_SCSISenseKeyTable[0x1][0x3][0x0].strErrorCondition = "Recovered Write error - write fault";
     m_SCSISenseKeyTable[0x1][0x9][0x0].strErrorCondition = "Track following error";
     m_SCSISenseKeyTable[0x1][0x0B][0x1].strErrorCondition = "Temperature warning";
     m_SCSISenseKeyTable[0x1][0x0C][0x1].strErrorCondition = "Recovered Write error with auto-realloc - reallocated";
     m_SCSISenseKeyTable[0x1][0x0C][0x3].strErrorCondition = "Recovered Write error - recommend reassign";
     m_SCSISenseKeyTable[0x1][0x12][0x1].strErrorCondition = "Recovered data without ECC using prev logical block ID";
     m_SCSISenseKeyTable[0x1][0x12][0x2].strErrorCondition = "Recovered data with ECC using prev logical block ID";
     m_SCSISenseKeyTable[0x1][0x14][0x1].strErrorCondition = "Recovered Record Not Found";
     m_SCSISenseKeyTable[0x1][0x16][0x0].strErrorCondition = "Recovered Write error - Data Sync Mark Error";
     m_SCSISenseKeyTable[0x1][0x16][0x1].strErrorCondition = "Recovered Write error - Data Sync Error - data rewritten";
     m_SCSISenseKeyTable[0x1][0x16][0x2].strErrorCondition = "Recovered Write error - Data Sync Error - recommend rewrite";
     m_SCSISenseKeyTable[0x1][0x16][0x3].strErrorCondition = "Recovered Write error - Data Sync Error - data auto-reallocated";
     m_SCSISenseKeyTable[0x1][0x16][0x4].strErrorCondition = "Recovered Write error - Data Sync Error - recommend reassignment";
     m_SCSISenseKeyTable[0x1][0x17][0x0].strErrorCondition = "Recovered data with no error correction applied";
     m_SCSISenseKeyTable[0x1][0x17][0x1].strErrorCondition = "Recovered Read error - with retries";
     m_SCSISenseKeyTable[0x1][0x17][0x2].strErrorCondition = "Recovered data using positive offset";
     m_SCSISenseKeyTable[0x1][0x17][0x3].strErrorCondition = "Recovered data using negative offset";
     m_SCSISenseKeyTable[0x1][0x17][0x5].strErrorCondition = "Recovered data using previous logical block ID";
     m_SCSISenseKeyTable[0x1][0x17][0x6].strErrorCondition = "Recovered Read error - without ECC, auto reallocated";
     m_SCSISenseKeyTable[0x1][0x17][0x7].strErrorCondition = "Recovered Read error - without ECC, recommend reassign";
     m_SCSISenseKeyTable[0x1][0x17][0x8].strErrorCondition = "Recovered Read error - without ECC, recommend rewrite";
     m_SCSISenseKeyTable[0x1][0x17][0x9].strErrorCondition = "Recovered Read error - without ECC, data rewritten";
     m_SCSISenseKeyTable[0x1][0x18][0x0].strErrorCondition = "Recovered Read error - with ECC";
     m_SCSISenseKeyTable[0x1][0x18][0x1].strErrorCondition = "Recovered data with ECC and retries";
     m_SCSISenseKeyTable[0x1][0x18][0x2].strErrorCondition = "Recovered Read error - with ECC, auto reallocated";
     m_SCSISenseKeyTable[0x1][0x18][0x5].strErrorCondition = "Recovered Read error - with ECC, recommend reassign";
     m_SCSISenseKeyTable[0x1][0x18][0x6].strErrorCondition = "Recovered data using ECC and offsets";
     m_SCSISenseKeyTable[0x1][0x18][0x7].strErrorCondition = "Recovered Read error - with ECC, data rewritten";
     m_SCSISenseKeyTable[0x1][0x1C][0x0].strErrorCondition = "Defect List not found";
     m_SCSISenseKeyTable[0x1][0x1C][0x1].strErrorCondition = "Primary defect list not found";
     m_SCSISenseKeyTable[0x1][0x1C][0x2].strErrorCondition = "Grown defect list not found";
     m_SCSISenseKeyTable[0x1][0x1F][0x0].strErrorCondition = "Partial defect list transferred";
     m_SCSISenseKeyTable[0x1][0x44][0x0].strErrorCondition = "Internal target failure";
     m_SCSISenseKeyTable[0x1][0x5D][0x0].strErrorCondition = "PFA threshold reached";
     m_SCSISenseKeyTable[0x2][0x4][0x0].strErrorCondition = "Not Ready - Cause not reportable.";
     m_SCSISenseKeyTable[0x2][0x4][0x1].strErrorCondition = "Not Ready - becoming ready";
     m_SCSISenseKeyTable[0x2][0x4][0x2].strErrorCondition = "Not Ready - need initialise command (start unit)";
     m_SCSISenseKeyTable[0x2][0x4][0x3].strErrorCondition = "Not Ready - manual intervention required";
     m_SCSISenseKeyTable[0x2][0x4][0x4].strErrorCondition = "Not Ready - format in progress";
     m_SCSISenseKeyTable[0x2][0x4][0x9].strErrorCondition = "Not Ready - self-test in progress";
     m_SCSISenseKeyTable[0x2][0x31][0x0].strErrorCondition = "Not Ready - medium format corrupted";
     m_SCSISenseKeyTable[0x2][0x31][0x1].strErrorCondition = "Not Ready - format command failed";
     m_SCSISenseKeyTable[0x2][0x35][0x2].strErrorCondition = "Not Ready - enclosure services unavailable";
     m_SCSISenseKeyTable[0x2][0x3A][0x0].strErrorCondition = "Not Ready - medium not present";
     m_SCSISenseKeyTable[0x2][0x3A][0x1].strErrorCondition = "Not Ready - medium not present - tray closed";
     m_SCSISenseKeyTable[0x2][0x3A][0x2].strErrorCondition = "Not Ready - medium not present - tray open";
     m_SCSISenseKeyTable[0x2][0x4C][0x0].strErrorCondition = "Diagnostic Failure - config not loaded";
     m_SCSISenseKeyTable[0x3][0x2][0x0].strErrorCondition = "Medium Error - No Seek Complete";
     m_SCSISenseKeyTable[0x3][0x3][0x0].strErrorCondition = "Medium Error - write fault";
     m_SCSISenseKeyTable[0x3][0x10][0x0].strErrorCondition = "Medium Error - ID CRC error";
     m_SCSISenseKeyTable[0x3][0x11][0x0].strErrorCondition = "Medium Error - unrecovered read error";
     m_SCSISenseKeyTable[0x3][0x11][0x1].strErrorCondition = "Medium Error - read retries exhausted";
     m_SCSISenseKeyTable[0x3][0x11][0x2].strErrorCondition = "Medium Error - error too long to correct";
     m_SCSISenseKeyTable[0x3][0x11][0x4].strErrorCondition = "Medium Error - unrecovered read error - auto re-alloc failed";
     m_SCSISenseKeyTable[0x3][0x11][0x0B].strErrorCondition = "Medium Error - unrecovered read error - recommend reassign";
     m_SCSISenseKeyTable[0x3][0x14][0x1].strErrorCondition = "Medium Error - record not found";
     m_SCSISenseKeyTable[0x3][0x16][0x0].strErrorCondition = "Medium Error - Data Sync Mark error";
     m_SCSISenseKeyTable[0x3][0x16][0x4].strErrorCondition = "Medium Error - Data Sync Error - recommend reassign";
     m_SCSISenseKeyTable[0x3][0x19][0x0].strErrorCondition = "Medium Error - defect list error";
     m_SCSISenseKeyTable[0x3][0x19][0x1].strErrorCondition = "Medium Error - defect list not available";
     m_SCSISenseKeyTable[0x3][0x19][0x2].strErrorCondition = "Medium Error - defect list error in primary list";
     m_SCSISenseKeyTable[0x3][0x19][0x3].strErrorCondition = "Medium Error - defect list error in grown list";
     m_SCSISenseKeyTable[0x3][0x19][0x0E].strErrorCondition = "Medium Error - fewer than 50% defect list copies";
     m_SCSISenseKeyTable[0x3][0x31][0x0].strErrorCondition = "Medium Error - medium format corrupted";
     m_SCSISenseKeyTable[0x3][0x31][0x1].strErrorCondition = "Medium Error - format command failed";
     m_SCSISenseKeyTable[0x4][0x1][0x0].strErrorCondition = "Hardware Error - no index or sector";
     m_SCSISenseKeyTable[0x4][0x2][0x0].strErrorCondition = "Hardware Error - no seek complete";
     m_SCSISenseKeyTable[0x4][0x3][0x0].strErrorCondition = "Hardware Error - write fault";
     m_SCSISenseKeyTable[0x4][0x9][0x0].strErrorCondition = "Hardware Error - track following error";
     m_SCSISenseKeyTable[0x4][0x11][0x0].strErrorCondition = "Hardware Error - unrecovered read error in reserved area";
     m_SCSISenseKeyTable[0x4][0x15][0x1].strErrorCondition = "Hardware Error - Mechanical positioning error";
     m_SCSISenseKeyTable[0x4][0x16][0x0].strErrorCondition = "Hardware Error - Data Sync Mark error in reserved area";
     m_SCSISenseKeyTable[0x4][0x19][0x0].strErrorCondition = "Hardware Error - defect list error";
     m_SCSISenseKeyTable[0x4][0x19][0x2].strErrorCondition = "Hardware Error - defect list error in Primary List";
     m_SCSISenseKeyTable[0x4][0x19][0x3].strErrorCondition = "Hardware Error - defect list error in Grown List";
     m_SCSISenseKeyTable[0x4][0x31][0x0].strErrorCondition = "Hardware Error - reassign failed";
     m_SCSISenseKeyTable[0x4][0x32][0x0].strErrorCondition = "Hardware Error - no defect spare available";
     m_SCSISenseKeyTable[0x4][0x35][0x1].strErrorCondition = "Hardware Error - unsupported enclosure function";
     m_SCSISenseKeyTable[0x4][0x35][0x2].strErrorCondition = "Hardware Error - enclosure services unavailable";
     m_SCSISenseKeyTable[0x4][0x35][0x3].strErrorCondition = "Hardware Error - enclosure services transfer failure";
     m_SCSISenseKeyTable[0x4][0x35][0x4].strErrorCondition = "Hardware Error - enclosure services refused";
     m_SCSISenseKeyTable[0x4][0x35][0x5].strErrorCondition = "Hardware Error - enclosure services checksum error";
     m_SCSISenseKeyTable[0x4][0x3E][0x3].strErrorCondition = "Hardware Error - self-test failed";
     m_SCSISenseKeyTable[0x4][0x3E][0x4].strErrorCondition = "Hardware Error - unable to update self-test";
     m_SCSISenseKeyTable[0x4][0x44][0x0].strErrorCondition = "Hardware Error - internal target failure";
     m_SCSISenseKeyTable[0x5][0x1A][0x0].strErrorCondition = "Illegal Request - parm list length error";
     m_SCSISenseKeyTable[0x5][0x20][0x0].strErrorCondition = "Illegal Request - invalid/unsupported command code";
     m_SCSISenseKeyTable[0x5][0x21][0x0].strErrorCondition = "Illegal Request - LBA out of range";
     m_SCSISenseKeyTable[0x5][0x24][0x0].strErrorCondition = "Illegal Request - invalid field in CDB (Command Descriptor Block)";
     m_SCSISenseKeyTable[0x5][0x25][0x0].strErrorCondition = "Illegal Request - invalid LUN";
     m_SCSISenseKeyTable[0x5][0x26][0x0].strErrorCondition = "Illegal Request - invalid fields in parm list";
     m_SCSISenseKeyTable[0x5][0x26][0x1].strErrorCondition = "Illegal Request - parameter not supported";
     m_SCSISenseKeyTable[0x5][0x26][0x2].strErrorCondition = "Illegal Request - invalid parm value";
     m_SCSISenseKeyTable[0x5][0x26][0x3].strErrorCondition = "Illegal Request - invalid field parameter - threshold parameter";
     m_SCSISenseKeyTable[0x5][0x26][0x4].strErrorCondition = "Illegal Request - invalid release of persistent reservation";
     m_SCSISenseKeyTable[0x5][0x2C][0x0].strErrorCondition = "Illegal Request - command sequence error";
     m_SCSISenseKeyTable[0x5][0x35][0x1].strErrorCondition = "Illegal Request - unsupported enclosure function";
     m_SCSISenseKeyTable[0x5][0x49][0x0].strErrorCondition = "Illegal Request - invalid message";
     m_SCSISenseKeyTable[0x5][0x53][0x0].strErrorCondition = "Illegal Request - media load or eject failed";
     m_SCSISenseKeyTable[0x5][0x53][0x1].strErrorCondition = "Illegal Request - unload tape failure";
     m_SCSISenseKeyTable[0x5][0x53][0x2].strErrorCondition = "Illegal Request - medium removal prevented";
     m_SCSISenseKeyTable[0x5][0x55][0x0].strErrorCondition = "Illegal Request - system resource failure";
     m_SCSISenseKeyTable[0x5][0x55][0x1].strErrorCondition = "Illegal Request - system buffer full";
     m_SCSISenseKeyTable[0x5][0x55][0x4].strErrorCondition = "Illegal Request - Insufficient Registration Resources";
     m_SCSISenseKeyTable[0x6][0x28][0x0].strErrorCondition = "Unit Attention - not-ready to ready transition (format complete)";
     m_SCSISenseKeyTable[0x6][0x29][0x0].strErrorCondition = "Unit Attention - POR or device reset occurred";
     m_SCSISenseKeyTable[0x6][0x29][0x1].strErrorCondition = "Unit Attention - POR occurred";
     m_SCSISenseKeyTable[0x6][0x29][0x2].strErrorCondition = "Unit Attention - SCSI bus reset occurred";
     m_SCSISenseKeyTable[0x6][0x29][0x3].strErrorCondition = "Unit Attention - TARGET RESET occurred";
     m_SCSISenseKeyTable[0x6][0x29][0x4].strErrorCondition = "Unit Attention - self-initiated-reset occurred";
     m_SCSISenseKeyTable[0x6][0x29][0x5].strErrorCondition = "Unit Attention - transceiver mode change to SE";
     m_SCSISenseKeyTable[0x6][0x29][0x6].strErrorCondition = "Unit Attention - transceiver mode change to LVD";
     m_SCSISenseKeyTable[0x6][0x2A][0x0].strErrorCondition = "Unit Attention - parameters changed";
     m_SCSISenseKeyTable[0x6][0x2A][0x1].strErrorCondition = "Unit Attention - mode parameters changed";
     m_SCSISenseKeyTable[0x6][0x2A][0x2].strErrorCondition = "Unit Attention - log select parms changed";
     m_SCSISenseKeyTable[0x6][0x2A][0x3].strErrorCondition = "Unit Attention - Reservations pre-empted";
     m_SCSISenseKeyTable[0x6][0x2A][0x4].strErrorCondition = "Unit Attention - Reservations released";
     m_SCSISenseKeyTable[0x6][0x2A][0x5].strErrorCondition = "Unit Attention - Registrations pre-empted";
     m_SCSISenseKeyTable[0x6][0x2F][0x0].strErrorCondition = "Unit Attention - commands cleared by another initiator";
     m_SCSISenseKeyTable[0x6][0x3F][0x0].strErrorCondition = "Unit Attention - target operating conditions have changed";
     m_SCSISenseKeyTable[0x6][0x3F][0x1].strErrorCondition = "Unit Attention - microcode changed";
     m_SCSISenseKeyTable[0x6][0x3F][0x2].strErrorCondition = "Unit Attention - changed operating definition";
     m_SCSISenseKeyTable[0x6][0x3F][0x3].strErrorCondition = "Unit Attention - inquiry parameters changed";
     m_SCSISenseKeyTable[0x6][0x3F][0x5].strErrorCondition = "Unit Attention - device identifier changed";
     m_SCSISenseKeyTable[0x6][0x5D][0x0].strErrorCondition = "Unit Attention - PFA threshold reached";
     m_SCSISenseKeyTable[0x7][0x27][0x0].strErrorCondition = "Write Protect - command not allowed";
     m_SCSISenseKeyTable[0xB][0x0][0x0].strErrorCondition = "Aborted Command - no additional sense code";
     m_SCSISenseKeyTable[0xB][0x1B][0x0].strErrorCondition = "Aborted Command - sync data transfer error (extra ACK)";
     m_SCSISenseKeyTable[0xB][0x25][0x0].strErrorCondition = "Aborted Command - unsupported LUN";
     m_SCSISenseKeyTable[0xB][0x3F][0x0F].strErrorCondition = "Aborted Command - echo buffer overwritten";
     m_SCSISenseKeyTable[0xB][0x43][0x0].strErrorCondition = "Aborted Command - message reject error";
     m_SCSISenseKeyTable[0xB][0x44][0x0].strErrorCondition = "Aborted Command - internal target failure";
     m_SCSISenseKeyTable[0xB][0x45][0x0].strErrorCondition = "Aborted Command - Selection/Reselection failure";
     m_SCSISenseKeyTable[0xB][0x47][0x0].strErrorCondition = "Aborted Command - SCSI parity error";
     m_SCSISenseKeyTable[0xB][0x48][0x0].strErrorCondition = "Aborted Command - initiator-detected error message received";
     m_SCSISenseKeyTable[0xB][0x49][0x0].strErrorCondition = "Aborted Command - inappropriate/illegal message";
     m_SCSISenseKeyTable[0xB][0x55][0x3].strErrorCondition = "Aborted Command - insufficient resources";
     m_SCSISenseKeyTable[0xB][0x4B][0x0].strErrorCondition = "Aborted Command - data phase error";
     m_SCSISenseKeyTable[0xB][0x4E][0x0].strErrorCondition = "Aborted Command - overlapped commands attempted";
     m_SCSISenseKeyTable[0xB][0x4F][0x0].strErrorCondition = "Aborted Command - due to loop initialisation";
     m_SCSISenseKeyTable[0xE][0x1D][0x0].strErrorCondition = "Miscompare - during verify byte check operation";
}
